import logging
import warnings
from typing import Optional

logger = logging.getLogger("momentosdk")


""" info('some %s stuff', 'information') """
info = logger.info


""" debug('some %s stuff', 'debug') """
debug = logger.debug


def add_logging_level(level_name: str, level_num: int, method_name: Optional[str] = None) -> bool:
    """Comprehensively adds a new logging level to the `logging` module and the currently configured logging class.

    `level_name` becomes an attribute of the `logging` module with the value
    `level_num`. `method_name` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `method_name` is not specified, `level_name.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present.

    Example::
    -------
    >>> add_logging_level('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name) or hasattr(logging, method_name) or hasattr(logging.getLoggerClass(), method_name):
        # Do nothing if the logging level already exists. Since this affects logging globally we can't know if the user
        # or another dependency has already added the same level.
        warnings.warn(
            "Logging level {level} or logging method {method} already defined. Skipping.".format(
                level=level_name, method=method_name
            )
        )
        return False

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):  # type: ignore[no-untyped-def]
        if self.isEnabledFor(level_num):  # type: ignore[misc]
            self._log(level_num, message, args, **kwargs)  # type: ignore[misc]

    def logToRoot(message, *args, **kwargs):  # type: ignore[no-untyped-def]
        logging.log(level_num, message, *args, **kwargs)  # type: ignore[misc]

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, logForLevel)  # type: ignore[misc]
    setattr(logging, method_name, logToRoot)  # type: ignore[misc]

    return True


TRACE = logging.DEBUG - 5


def initialize_momento_logging() -> None:
    add_logging_level("TRACE", TRACE)



from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheListLengthResponse(CacheResponse):
    """Response type for a `list_length` request.

    Its subtypes are:
    - `CacheListLength.Hit`
    - `CacheListLength.Miss`
    - `CacheListLength.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheListLength(ABC):
    """Groups all `CacheListLengthResponse` derived types under a common namespace."""

    @dataclass
    class Hit(CacheListLengthResponse):
        """Indicates the list exists and its length was fetched."""

        length: int
        """The number of values in the list."""

    class Miss(CacheListLengthResponse):
        """Indicates the list does not exist."""

    class Error(CacheListLengthResponse, ErrorResponseMixin):
        """Indicates an error occured in the request.

        This includes
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



import enum
from dataclasses import dataclass
from typing import Optional

import grpc


@enum.unique
class MomentoErrorCode(enum.Enum):
    """A list of all available Momento error codes.

    These can be used to check for specific types of errors on a failure response.
    """

    INVALID_ARGUMENT_ERROR = (1,)
    """Invalid argument passed to Momento client"""
    UNKNOWN_SERVICE_ERROR = (2,)
    """Service returned an unknown response"""
    ALREADY_EXISTS_ERROR = (3,)
    """Cache with specified name already exists"""
    NOT_FOUND_ERROR = (4,)
    """Cache with specified name doesn't exist"""
    INTERNAL_SERVER_ERROR = (5,)
    """An unexpected error occurred while trying to fulfill the request"""
    PERMISSION_ERROR = (6,)
    """Insufficient permissions to perform operation"""
    AUTHENTICATION_ERROR = (7,)
    """Invalid authentication credentials to connect to cache service"""
    CANCELLED_ERROR = (8,)
    """Request was cancelled by the server"""
    LIMIT_EXCEEDED_ERROR = (9,)
    """Request rate exceeded the limits for the account"""
    BAD_REQUEST_ERROR = (10,)
    """Request was invalid"""
    TIMEOUT_ERROR = (11,)
    """Client's configured timeout was exceeded"""
    SERVER_UNAVAILABLE = (12,)
    """Server was unable to handle the request"""
    CLIENT_RESOURCE_EXHAUSTED = (13,)
    """A client resource most likely memory was exhausted"""
    FAILED_PRECONDITION_ERROR = (14,)
    """System is not in a state required for the operation's execution"""
    UNKNOWN_ERROR = 15
    """Unknown error has occurred"""


@dataclass
class MomentoGrpcErrorDetails:
    """Captures low-level information about an error, at the gRPC level.

    Hopefully this is only needed in rare cases, by Momento engineers, for debugging.
    """

    code: grpc.StatusCode
    """The gRPC status code of the error repsonse"""
    details: str
    """Detailed information about the error"""
    metadata: Optional[grpc.aio.Metadata] = None
    """Headers and other information about the error response"""
    # TODO need to reconcile synchronous metadata (above) with grpc.aio.Metadata


@dataclass
class MomentoErrorTransportDetails:
    """Container for low-level error information, including details from the transport layer."""

    grpc: MomentoGrpcErrorDetails
    """Low-level gRPC error details"""



from typing import Optional

from momento.errors import MomentoErrorCode, MomentoErrorTransportDetails


class SdkException(Exception):
    """Base class for all Momento client exceptions."""

    message: str
    """Exception message"""
    error_code: MomentoErrorCode
    """Enumeration of all possible Momento error types.  Should be used in
    code to distinguish between different types of errors.
    """
    transport_details: Optional[MomentoErrorTransportDetails] = None
    """Low-level error details, from the transport layer.  Hopefully only needed
    in rare cases, by Momento engineers, for debugging.
    """
    message_wrapper: str
    """Prefix with basic information about the error class; this will be appended
    with specific information about the individual error instance at runtime."""

    def __init__(
        self,
        message: str,
        error_code: MomentoErrorCode,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
        message_wrapper: str = "",
    ):
        super().__init__(message)
        self.message = message
        self.error_code = error_code
        self.transport_details = transport_details
        self.message_wrapper = message_wrapper

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"message={self.message!r}, "
            f"error_code={self.error_code!r}, "
            f"transport_details={self.transport_details!r}, "
            f"message_wrapper={self.message_wrapper!r})"
        )

    def __str__(self) -> str:
        return repr(self)


class AlreadyExistsException(SdkException):
    """Resource already exists."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.ALREADY_EXISTS_ERROR,
            transport_details,
            message_wrapper="A cache with the specified name already exists. To resolve this error, either delete the existing cache and make a new one, or use a different name",  # noqa: E501
        )


class AuthenticationException(SdkException):
    """Authentication token is not provided or is invalid."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.AUTHENTICATION_ERROR,
            transport_details,
            message_wrapper="Invalid authentication credentials to connect to cache service",
        )


class BadRequestException(SdkException):
    """Invalid parameters sent to Momento Services."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.BAD_REQUEST_ERROR,
            transport_details,
            message_wrapper="The request was invalid; please contact us at support@momentohq.com",
        )


class CancelledException(SdkException):
    """Operation was cancelled."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.CANCELLED_ERROR,
            transport_details,
            message_wrapper="The request was cancelled by the server; please contact us at support@momentohq.com",
        )


class FailedPreconditionException(SdkException):
    """The server did not meet the precondition to run a command.

    For example, calling `Increment` on a key that doesn't store
    a number.
    """

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.FAILED_PRECONDITION_ERROR,
            transport_details,
            message_wrapper="System is not in a state required for the operation's execution",
        )


class InternalServerException(SdkException):
    """Momento Service encountered an unexpected exception while trying to fulfill the request."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.INTERNAL_SERVER_ERROR,
            transport_details,
            message_wrapper="An unexpected error occurred while trying to fulfill the request; please contact us at support@momentohq.com",  # noqa: E501
        )


class InvalidArgumentException(SdkException):
    """SDK client-side validation failed."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.INVALID_ARGUMENT_ERROR,
            transport_details,
            message_wrapper="Invalid argument passed to Momento client",
        )


class LimitExceededException(SdkException):
    """Requested operation couldn't be completed because system limits were hit."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.LIMIT_EXCEEDED_ERROR,
            transport_details,
            message_wrapper="Request rate, bandwidth, or object size exceeded the limits for this account. To resolve this error, reduce your usage as appropriate or contact us at support@momentohq.com to request a limit increase",  # noqa: E501
        )


class NotFoundException(SdkException):
    """Requested resource or the resource on which an operation was requested doesn't exist."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.NOT_FOUND_ERROR,
            transport_details,
            message_wrapper="A cache with the specified name does not exist. To resolve this error, make sure you have created the cache before attempting to use it",  # noqa: E501
        )


class PermissionDeniedException(SdkException):
    """Insufficient permissions to execute an operation."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.PERMISSION_ERROR,
            transport_details,
            message_wrapper="Insufficient permissions to perform an operation on a cache",
        )


class ServerUnavailableException(SdkException):
    """Server was unable to handle the request."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.SERVER_UNAVAILABLE,
            transport_details,
            message_wrapper="The server was unable to handle the request; consider retrying. If the error persists, please contact us at support@momentohq.com",  # noqa: E501
        )


class TimeoutException(SdkException):
    """Requested operation did not complete in allotted time."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.TIMEOUT_ERROR,
            transport_details,
            message_wrapper="The client's configured timeout was exceeded; you may need to use a Configuration with more lenient timeouts",  # noqa: E501
        )


class UnknownException(SdkException):
    """Unhandled exception from CacheExceptionMapper."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message, MomentoErrorCode.UNKNOWN_ERROR, transport_details, message_wrapper="Unknown error has occurred"
        )


class UnknownServiceException(SdkException):
    """Service returned an unknown response."""

    def __init__(
        self,
        message: str,
        transport_details: Optional[MomentoErrorTransportDetails] = None,
    ):
        super().__init__(
            message,
            MomentoErrorCode.BAD_REQUEST_ERROR,
            transport_details,
            message_wrapper="Service returned an unknown response; please contact us at support@momentohq.com",
        )



from dataclasses import dataclass

import grpc


# The lack of type hints in the grpc code causes a lint failure in combination with the dataclass annotation
@dataclass  # type: ignore[misc]
class RetryableProps:
    grpc_status: grpc.StatusCode
    grpc_method: str
    attempt_number: int



import typing

# https://stackoverflow.com/questions/71889556/mypy-checking-typing-protocol-with-python-3-7-support
if typing.TYPE_CHECKING:
    from typing_extensions import Protocol
else:
    Protocol = object

from momento.errors import MomentoErrorCode, SdkException


class ValueStringMixin:
    """Renders `value_bytes` as a utf-8 string.

    Returns:
        str: the utf-8 encoding of the data
    """

    value_bytes: bytes

    @property
    def value_string(self) -> str:
        """Convert the bytes `value` to a UTF-8 string.

        Returns:
            str: UTF-8 representation of the `value`
        """
        return self.value_bytes.decode("utf-8")


class ErrorResponseMixin:
    _error: SdkException

    def __init__(self, _error: SdkException):
        self._error = _error

    @property
    def inner_exception(self) -> SdkException:
        """The SdkException object used to construct the response."""
        return self._error

    @property
    def error_code(self) -> MomentoErrorCode:
        """The `MomentoErrorCode` value for the particular error object."""
        return self._error.error_code

    @property
    def message(self) -> str:
        """An explanation of conditions that caused and potential ways to resolve the error."""
        return f"{self._error.message_wrapper}: {self._error.message}"

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__} {self.error_code}: {self.message}"



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheListFetchResponse(CacheResponse):
    """Response type for a `list_fetch` request.

    Its subtypes are:
    - `CacheListFetch.Hit`
    - `CacheListFetch.Miss`
    - `CacheListFetch.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheListFetch(ABC):
    """Groups all `CacheListFetchResponse` derived types under a common namespace."""

    @dataclass
    class Hit(CacheListFetchResponse):
        """Indicates the list exists and its values were fetched."""

        value_list_bytes: list[bytes]
        """The values for the fetched list, as bytes."""

        @property
        def value_list_string(self) -> list[str]:
            """The values for the fetched list, as utf-8 encoded strings.

            Returns:
                list[str]
            """
            return [v.decode("utf-8") for v in self.value_list_bytes]

    class Miss(CacheListFetchResponse):
        """Indicates the list does not exist."""

    class Error(CacheListFetchResponse, ErrorResponseMixin):
        """Indicates an error occured in the request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

import logging
import time
from typing import Callable, TypeVar

import grpc

from momento.retry import RetryableProps, RetryStrategy

RequestType = TypeVar("RequestType")
InterceptorCall = TypeVar("InterceptorCall")
ResponseType = TypeVar("ResponseType")

# TODO: This is very duplicative of the asyncio retry interceptor; we need to
# DRY these up, but for now I am prioritizing getting a fix out for a customer.

logger = logging.getLogger("retry-interceptor")


# TODO: We need to send retry count information to the server so that we
# will have some visibility into how often this is happening to customers:
# https://github.com/momentohq/client-sdk-javascript/issues/80
# TODO: we need to add backoff/jitter for the retries:
# https://github.com/momentohq/client-sdk-javascript/issues/81


class RetryInterceptor(grpc.UnaryUnaryClientInterceptor):
    def __init__(self, retry_strategy: RetryStrategy):
        self._retry_strategy = retry_strategy

    def intercept_unary_unary(
        self,
        continuation: Callable[[grpc.ClientCallDetails, RequestType], InterceptorCall],
        client_call_details: grpc.ClientCallDetails,
        request: RequestType,
    ) -> InterceptorCall | ResponseType:
        attempt_number = 1
        while True:
            call = continuation(client_call_details, request)
            response_code = call.code()  # type: ignore[attr-defined]  # noqa: F401

            retryTime = self._retry_strategy.determine_when_to_retry(
                RetryableProps(response_code, client_call_details.method, attempt_number)
            )

            if retryTime is None:
                return call

            attempt_number += 1
            time.sleep(retryTime)



from ._data_validation import (
    _as_bytes,
    _gen_dictionary_fields_as_bytes,
    _gen_dictionary_items_as_bytes,
    _gen_list_as_bytes,
    _gen_set_input_as_bytes,
    _validate_cache_name,
    _validate_dictionary_name,
    _validate_list_name,
    _validate_request_timeout,
    _validate_set_name,
    _validate_timedelta_ttl,
    _validate_ttl,
)



from abc import ABC

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheDictionarySetFieldResponse(CacheResponse):
    """Parent response type for a cache `dictionary_set_field` request.

    Its subtypes are:
    - `CacheDictionarySetField.Success`
    - `CacheDictionarySetField.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheDictionarySetField(ABC):
    """Groups all `CacheDictionarySetFieldResponse` derived types under a common namespace."""

    class Success(CacheDictionarySetFieldResponse):
        """Indicates the request was successful."""

    class Error(CacheDictionarySetFieldResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheSetFetchResponse(CacheResponse):
    """Parent response type for a `set_fetch` request.

    Its subtypes are:
    - `CacheSetFetch.Hit`
    - `CacheSetFetch.Miss`
    - `CacheSetFetch.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheSetFetch(ABC):
    """Groups all `CacheSetFetchResponse` derived types under a common namespace."""

    @dataclass
    class Hit(CacheSetFetchResponse):
        """Indicates the set exists and its values were fetched."""

        value_set_bytes: set[bytes]
        """The elements as a Python set.

        Use value_set_string to get the elements as a set.
        """

        @property
        def value_set_string(self) -> set[str]:
            """The elements of the set, as utf-8 encoded strings.

            Returns:
                TSetElementsOutputStr
            """
            return {v.decode("utf-8") for v in self.value_set_bytes}

    class Miss(CacheSetFetchResponse):
        """Indicates the set does not exist."""

    class Error(CacheSetFetchResponse, ErrorResponseMixin):
        """Indicates an error occured in the request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from typing import Any

import google

from momento.responses.response import ControlResponse

from ...mixins import ErrorResponseMixin


@dataclass
class SigningKey:
    """Signing keys returned from requesting list signing keys."""

    key_id: str
    """the ID of the signing key"""
    expires_at: datetime
    """when the key expires"""
    endpoint: str
    """endpoint of the signing key"""

    @staticmethod
    def from_grpc_response(grpc_listed_signing_key: Any, endpoint: str) -> SigningKey:  # type: ignore[misc]
        key_id: str = grpc_listed_signing_key.key_id
        expires_at: datetime = datetime.fromtimestamp(grpc_listed_signing_key.expires_at)
        return SigningKey(key_id, expires_at, endpoint)


class ListSigningKeysResponse(ControlResponse):
    """Parent response type for a cache `list_signing_key` request.

    Its subtypes are:
    - `ListSigningKeys.Success`
    - `ListSigningKeys.Error`

    See `CacheClient` for how to work with responses.
    """


class ListSigningKeys(ABC):
    """Groups all `ListSigningKeysResponse` derived types under a common namespace."""

    @dataclass
    class Success(ListSigningKeysResponse):
        """The response from listing signing keys."""

        signing_keys: list[SigningKey]
        """all signing keys in this page"""

        @staticmethod
        def from_grpc_response(  # type:ignore[misc]
            grpc_list_signing_keys_response: google.protobuf.message.Message, endpoint: str
        ) -> ListSigningKeys.Success:
            """Creates a ListSigningKeysResponse from a grpc response.

            Args:
                grpc_list_signing_keys_response (google.protobuf.message.Message):
                endpoint (str): _description_

            Returns:
                ListSigningKeysResponse
            """
            signing_keys: list[SigningKey] = [
                SigningKey.from_grpc_response(signing_key, endpoint)
                for signing_key in grpc_list_signing_keys_response.signing_key
            ]
            return ListSigningKeys.Success(signing_keys)

    class Error(ListSigningKeysResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from datetime import timedelta

import grpc
from momento_wire_types.controlclient_pb2 import (
    _CreateCacheRequest,
    _CreateSigningKeyRequest,
    _DeleteCacheRequest,
    _ListCachesRequest,
    _ListSigningKeysRequest,
    _RevokeSigningKeyRequest,
)
from momento_wire_types.controlclient_pb2_grpc import ScsControlStub

from momento import logs
from momento.auth import CredentialProvider
from momento.config import Configuration
from momento.errors import convert_error
from momento.internal._utilities import _validate_cache_name, _validate_ttl
from momento.internal.synchronous._scs_grpc_manager import _ControlGrpcManager
from momento.responses import (
    CreateCache,
    CreateCacheResponse,
    CreateSigningKey,
    CreateSigningKeyResponse,
    DeleteCache,
    DeleteCacheResponse,
    ListCaches,
    ListCachesResponse,
    ListSigningKeys,
    ListSigningKeysResponse,
    RevokeSigningKey,
    RevokeSigningKeyResponse,
)

_DEADLINE_SECONDS = 60.0  # 1 minute


class _ScsControlClient:
    """Momento Internal."""

    def __init__(self, configuration: Configuration, credential_provider: CredentialProvider):
        endpoint = credential_provider.control_endpoint
        self._logger = logs.logger
        self._logger.debug("Simple cache control client instantiated with endpoint: %s", endpoint)
        self._grpc_manager = _ControlGrpcManager(configuration, credential_provider)
        self._endpoint = endpoint

    @property
    def endpoint(self) -> str:
        return self._endpoint

    def create_cache(self, cache_name: str) -> CreateCacheResponse:
        try:
            self._logger.info(f"Creating cache with name: {cache_name}")
            _validate_cache_name(cache_name)
            request = _CreateCacheRequest()
            request.cache_name = cache_name
            self._build_stub().CreateCache(request, timeout=_DEADLINE_SECONDS)
        except Exception as e:
            self._logger.debug("Failed to create cache: %s with exception: %s", cache_name, e)
            if isinstance(e, grpc.RpcError) and e.code() == grpc.StatusCode.ALREADY_EXISTS:
                return CreateCache.CacheAlreadyExists()
            return CreateCache.Error(convert_error(e))
        return CreateCache.Success()

    def delete_cache(self, cache_name: str) -> DeleteCacheResponse:
        try:
            self._logger.info(f"Deleting cache with name: {cache_name}")
            _validate_cache_name(cache_name)
            request = _DeleteCacheRequest()
            request.cache_name = cache_name
            self._build_stub().DeleteCache(request, timeout=_DEADLINE_SECONDS)
        except Exception as e:
            self._logger.debug("Failed to delete cache: %s with exception: %s", cache_name, e)
            return DeleteCache.Error(convert_error(e))
        return DeleteCache.Success()

    def list_caches(self) -> ListCachesResponse:
        try:
            list_caches_request = _ListCachesRequest()
            list_caches_request.next_token = ""
            response = self._build_stub().ListCaches(list_caches_request, timeout=_DEADLINE_SECONDS)
            return ListCaches.Success.from_grpc_response(response)
        except Exception as e:
            return ListCaches.Error(convert_error(e))

    def create_signing_key(self, ttl: timedelta, endpoint: str) -> CreateSigningKeyResponse:
        try:
            self._logger.info(f"Creating signing key with ttl={ttl!r} and endpoint={endpoint!r}")
            _validate_ttl(ttl)
            ttl_minutes = round(ttl.total_seconds() / 60)
            self._logger.info(f"Creating signing key with ttl (in minutes): {ttl_minutes}")
            create_signing_key_request = _CreateSigningKeyRequest()
            create_signing_key_request.ttl_minutes = ttl_minutes
            response = self._build_stub().CreateSigningKey(create_signing_key_request, timeout=_DEADLINE_SECONDS)
            return CreateSigningKey.Success.from_grpc_response(response, endpoint)
        except Exception as e:
            self._logger.warning(f"Failed to create signing key with exception: {e}")
            return CreateSigningKey.Error(convert_error(e))

    def revoke_signing_key(self, key_id: str) -> RevokeSigningKeyResponse:
        try:
            self._logger.info(f"Revoking signing key with key_id {key_id}")
            request = _RevokeSigningKeyRequest()
            request.key_id = key_id
            self._build_stub().RevokeSigningKey(request, timeout=_DEADLINE_SECONDS)
            return RevokeSigningKey.Success()
        except Exception as e:
            self._logger.warning(f"Failed to revoke signing key with key_id {key_id} exception: {e}")
            return RevokeSigningKey.Error(convert_error(e))

    def list_signing_keys(self, endpoint: str) -> ListSigningKeysResponse:
        try:
            self._logger.info("List signing keys")
            list_signing_keys_request = _ListSigningKeysRequest()
            list_signing_keys_request.next_token = ""
            response = self._build_stub().ListSigningKeys(list_signing_keys_request, timeout=_DEADLINE_SECONDS)
            return ListSigningKeys.Success.from_grpc_response(
                response,
                endpoint,
            )
        except Exception as e:
            self._logger.warning(f"Failed to list signing keys with exception: {e}")
            return ListSigningKeys.Error(convert_error(e))

    def _build_stub(self) -> ScsControlStub:
        return self._grpc_manager.stub()

    def close(self) -> None:
        self._grpc_manager.close()



from abc import ABC

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheSetIfNotExistsResponse(CacheResponse):
    """Parent response type for a cache `set_if_not_exists` request.

    Its subtypes are:
    - `CacheSetIfNotExists.Stored`
    - `CacheSetIfNotExists.NotStored`
    - `CacheSetIfNotExists.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheSetIfNotExists(ABC):
    """Groups all `CacheSetIfNotExistsResponse` derived types under a common namespace."""

    class Stored(CacheSetIfNotExistsResponse):
        """Indicates the key did not exist and the value was set."""

    class NotStored(CacheSetIfNotExistsResponse):
        """Indicates the key existed and no value was set."""

    class Error(CacheSetIfNotExistsResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



import logging
from typing import Optional

from .default_eligibility_strategy import DefaultEligibilityStrategy
from .eligibility_strategy import EligibilityStrategy
from .retry_strategy import RetryStrategy
from .retryable_props import RetryableProps

logger = logging.getLogger("fixed-count-retry-strategy")


class FixedCountRetryStrategy(RetryStrategy):
    def __init__(
        self, *, max_attempts: int = 3, eligibility_strategy: DefaultEligibilityStrategy = DefaultEligibilityStrategy()
    ):
        self._eligibility_strategy: EligibilityStrategy = eligibility_strategy
        self._max_attempts: int = max_attempts

    def determine_when_to_retry(self, props: RetryableProps) -> Optional[float]:
        """Determines whether a grpc call can be retried and how long to wait before that retry.

        Args:
            props (RetryableProps): Information about the grpc call, its last invocation, and how many times the call
            has been made.

        :Returns
            The time in seconds before the next retry should occur or None if no retry should be attempted.
        """
        if self._eligibility_strategy.is_eligible_for_retry(props) is False:
            logger.debug(
                "Request path: %s; retryable status code: %s. Request is not retryable.",
                props.grpc_method,
                props.grpc_status,  # type: ignore[misc]
            )
            return None

        if props.attempt_number > self._max_attempts:
            logger.debug(
                "Request path: %s; retryable status code: %s; number of attempts (%i) "
                "has exceeded max (%i); not retrying.",
                props.grpc_method,
                props.grpc_status,  # type: ignore[misc]
                props.attempt_number,
                self._max_attempts,
            )
            return None

        logger.debug(
            "Request path: %s; retryable status code: %s; number of attempts (%i) " "is not above max (%i); retrying.",
            props.grpc_method,
            props.grpc_status,  # type: ignore[misc]
            props.attempt_number,
            self._max_attempts,
        )
        return 0.0



from .error_details import (
    MomentoErrorCode,
    MomentoErrorTransportDetails,
    MomentoGrpcErrorDetails,
)
from .exceptions import (
    AlreadyExistsException,
    AuthenticationException,
    BadRequestException,
    CancelledException,
    FailedPreconditionException,
    InternalServerException,
    InvalidArgumentException,
    LimitExceededException,
    NotFoundException,
    PermissionDeniedException,
    SdkException,
    ServerUnavailableException,
    TimeoutException,
    UnknownException,
    UnknownServiceException,
)

# NB: since this module imports from sibling modules, it must be at the bottom
# to avoid circular imports
from .error_converter import convert_error

__all__ = [
    "MomentoErrorCode",
    "MomentoErrorTransportDetails",
    "MomentoGrpcErrorDetails",
    "AlreadyExistsException",
    "AuthenticationException",
    "BadRequestException",
    "CancelledException",
    "FailedPreconditionException",
    "InternalServerException",
    "InvalidArgumentException",
    "LimitExceededException",
    "NotFoundException",
    "PermissionDeniedException",
    "SdkException",
    "ServerUnavailableException",
    "TimeoutException",
    "UnknownException",
    "UnknownServiceException",
    "convert_error",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import timedelta


class GrpcConfiguration(ABC):
    @abstractmethod
    def get_deadline(self) -> timedelta:
        pass

    @abstractmethod
    def with_deadline(self, deadline: timedelta) -> GrpcConfiguration:
        pass



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import List

from momento_wire_types.controlclient_pb2 import _ListCachesResponse

from momento.responses.response import ControlResponse

from ...mixins import ErrorResponseMixin


class ListCachesResponse(ControlResponse):
    """Parent response type for a list caches request.

    The response object is resolved to a type-safe object of one of
    the following subtypes:
    - `ListCaches.Success`
    - `ListCaches.Error`

    Pattern matching can be used to operate on the appropriate subtype.
    For example, in python 3.10+::

        match response:
            case ListCaches.Success():
                ...
            case ListCaches.Error():
                ...
            case _:
                # Shouldn't happen

    or equivalently in earlier versions of python::

        if isinstance(response, ListCaches.Success):
            ...
        elif isinstance(response, ListCaches.Error):
            ...
        else:
            # Shouldn't happen
    """


@dataclass
class CacheInfo:
    """Contains a Momento cache's name."""

    name: str
    """Holds the name of the cache."""


class ListCaches(ABC):
    """Groups all `ListCachesResponse` derived types under a common namespace."""

    @dataclass
    class Success(ListCachesResponse):
        """Indicates the request was successful."""

        caches: List[CacheInfo]
        """The list of caches available to the user."""

        @staticmethod
        def from_grpc_response(grpc_list_cache_response: _ListCachesResponse) -> ListCaches.Success:  # type: ignore[misc] # noqa: E501
            """Initializes ListCacheResponse to handle list cache response.

            Args:
                grpc_list_cache_response: Protobuf based response returned by Scs.
            """
            caches = [CacheInfo(cache.cache_name) for cache in grpc_list_cache_response.cache]  # type: ignore[misc]
            return ListCaches.Success(caches=caches)

    class Error(ListCachesResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from abc import ABC

from momento.responses.response import ControlResponse

from ...mixins import ErrorResponseMixin


class DeleteCacheResponse(ControlResponse):
    """Parent response type for a delete cache request.

    The response object is resolved to a type-safe object of one of
    the following subtypes:
    - `DeleteCache.Success`
    - `DeleteCache.Error`

    Pattern matching can be used to operate on the appropriate subtype.
    For example, in python 3.10+::

        match response:
            case DeleteCache.Success():
                ...
            case DeleteCache.Error():
                ...
            case _:
                # Shouldn't happen

    or equivalently in earlier versions of python::

        if isinstance(response, DeleteCache.Success):
            ...
        elif isinstance(response, DeleteCache.Error):
            ...
        else:
            # Shouldn't happen
    """


class DeleteCache(ABC):
    """Groups all `DeleteCacheResponse` derived types under a common namespace."""

    class Success(DeleteCacheResponse):
        """Indicates the request was successful."""

    class Error(DeleteCacheResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin, ValueStringMixin
from ...response import CacheResponse


class CacheDictionaryGetFieldResponse(CacheResponse):
    """Parent response type for a cache `dictionary_get_field` request.

    Its subtypes are:
    - `CacheDictionaryGetField.Hit`
    - `CacheDictionaryGetField.Miss`
    - `CacheDictionaryGetField.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheDictionaryGetField(ABC):
    """Groups all `CacheDictionaryGetFieldResponse` derived types under a common namespace."""

    @dataclass
    class Hit(CacheDictionaryGetFieldResponse, ValueStringMixin):
        """Contains the result of a cache hit."""

        value_bytes: bytes
        """The value returned from the cache for the specified field. Use the
        `value_string` property to access the value as a string."""

        field_bytes: bytes
        """The dictionary field that was queried."""

        @property
        def field_string(self) -> str:
            """Convert the bytes `field` to a UTF-8 string.

            Returns:
                str: UTF-8 representation of the `field`
            """
            return self.field_bytes.decode("utf-8")

    class Miss(CacheDictionaryGetFieldResponse):
        """Indicates the dictionary or dictionary field does not exist."""

    class Error(CacheDictionaryGetFieldResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheListPushFrontResponse(CacheResponse):
    """Response type for a `list_push_front` request.

    Its subtypes are:
    - `CacheListPushFront.Success`
    - `CacheListPushFront.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheListPushFront(ABC):
    """Groups all `CacheListPushFrontResponse` derived types under a common namespace."""

    @dataclass
    class Success(CacheListPushFrontResponse):
        """Indicates the push was successful."""

        list_length: int
        """The number of values in the list after this push"""

    class Error(CacheListPushFrontResponse, ErrorResponseMixin):
        """Indicates an error occured in the request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse
from .get_field import CacheDictionaryGetField, CacheDictionaryGetFieldResponse


class CacheDictionaryGetFieldsResponse(CacheResponse):
    """Parent response type for a cache `dictionary_get_fields` request.

    Its subtypes are:
    - `CacheDictionaryGetFields.Hit`
    - `CacheDictionaryGetFields.Miss`
    - `CacheDictionaryGetFields.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheDictionaryGetFields(ABC):
    """Groups all `CacheDictionaryGetFieldsResponse` derived types under a common namespace."""

    @dataclass
    class Hit(CacheDictionaryGetFieldsResponse):
        """Contains the result of a cache hit."""

        responses: list[CacheDictionaryGetFieldResponse]
        """The value returned from the cache for the specified field. Use the
        `value_string` property to access the value as a string."""

        @property
        def value_dictionary_bytes_bytes(self) -> dict[bytes, bytes]:
            """The items for the dictionary fields, as a mapping from bytes to bytes.

            Returns:
                dict[bytes, bytes]
            """
            return {
                response.field_bytes: response.value_bytes
                for response in self.responses
                if isinstance(response, CacheDictionaryGetField.Hit)
            }

        @property
        def value_dictionary_bytes_string(self) -> dict[bytes, str]:
            """The items for the dictionary fields, as a mapping from bytes to utf-8 encoded strings.

            Returns:
                dict[bytes, str]
            """
            return {
                response.field_bytes: response.value_string
                for response in self.responses
                if isinstance(response, CacheDictionaryGetField.Hit)
            }

        @property
        def value_dictionary_string_bytes(self) -> dict[str, bytes]:
            """The items for the dictionary fields, as a mapping from utf-8 encoded strings to bytes.

            Returns:
                dict[str, bytes]
            """
            return {
                response.field_string: response.value_bytes
                for response in self.responses
                if isinstance(response, CacheDictionaryGetField.Hit)
            }

        @property
        def value_dictionary_string_string(self) -> dict[str, str]:
            """The items for the dictionary fields, as a mapping from utf-8 encoded strings to utf-8 encoded strings.

            Returns:
                dict[str, str]
            """
            return {
                response.field_string: response.value_string
                for response in self.responses
                if isinstance(response, CacheDictionaryGetField.Hit)
            }

    class Miss(CacheDictionaryGetFieldsResponse):
        """Indicates the dictionary does not exist."""

    class Error(CacheDictionaryGetFieldsResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

from typing import Iterable, List, Mapping, Set, Union

TCacheName = str

TMomentoValue = Union[str, bytes]

# Scalar Types
TScalarKey = Union[str, bytes]
TScalarValue = TMomentoValue

# Collections
TCollectionName = str
TCollectionValue = Union[str, bytes]

# Dictionary Types
TDictionaryName = TCollectionName
TDictionaryField = Union[str, bytes]
TDictionaryValue = TMomentoValue
TDictionaryFields = Iterable[TDictionaryField]
TDictionaryItems = Union[
    Mapping[TDictionaryField, TDictionaryValue],
    # Mapping isn't covariant so we have to list out the types here
    Mapping[bytes, bytes],
    Mapping[bytes, str],
    Mapping[str, bytes],
    Mapping[str, str],
]

# List Types
TListName = TCollectionName
TListValue = TMomentoValue
TListValuesInputBytes = Iterable[bytes]
TListValuesInputStr = Iterable[str]
TListValuesInput = Iterable[TListValue]
TListValuesOutputBytes = List[bytes]
TListValuesOutputStr = List[str]
TLIstValuesOutput = Union[TListValuesOutputBytes, TListValuesOutputStr]

# Set Types
TSetName = TCollectionName
TSetElement = TMomentoValue
TSetElementsInputBytes = Iterable[bytes]
TSetElementsInputStr = Iterable[str]
TSetElementsInput = Iterable[TSetElement]
TSetElementsOutputStr = Set[str]
TSetElementsOutputBytes = Set[bytes]
TSetElementsOutput = Union[TSetElementsOutputBytes, TSetElementsOutputStr]



from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin, ValueStringMixin
from ...response import CacheResponse


class CacheListPopFrontResponse(CacheResponse):
    """Parent response type for a `list_pop_front` request.

    Its subtypes are:
    - `CacheListPopFront.Hit`
    - `CacheListPopFront.Miss`
    - `CacheListPopFront.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheListPopFront(ABC):
    """Groups all `CacheListPopFront` derived types under a common namespace."""

    @dataclass
    class Hit(CacheListPopFrontResponse, ValueStringMixin):
        """Indicates the request was successful."""

        value_bytes: bytes
        """The value popped from the list. Use the `value_string` property to access the value as a string."""

    class Miss(CacheListPopFrontResponse):
        """Indicates the list does not exist."""

    class Error(CacheListPopFrontResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

from abc import ABC
from typing import Any, no_type_check


class Response(ABC):
    """Parent of all responses."""

    # We need to work with Any.
    @no_type_check
    def _render_as_str(self, max_value_length: int = 32, max_collection_length: int = 5) -> str:
        class_name = self.__class__.__qualname__
        attributes = vars(self)

        if not attributes:
            return f"{class_name}()"

        message_parts = []
        for attribute, value in attributes.items():
            appended = False

            # To truncate a collection, we must render it as a string all at once to account for the ellipsis
            if type(value) == list:
                if len(value) > max_collection_length:
                    message_parts.append(
                        Response._display_list(attribute, value, max_value_length, max_collection_length)
                    )
                    appended = True
                else:
                    value = [Response._truncate_value(v_i, max_value_length) for v_i in value]
            elif type(value) == set:
                if len(value) > max_collection_length:
                    message_parts.append(
                        Response._display_set(attribute, value, max_value_length, max_collection_length)
                    )
                    appended = True
                else:
                    value = {Response._truncate_value(v_i, max_value_length) for v_i in value}
            elif type(value) == dict:
                if len(value) > max_collection_length:
                    message_parts.append(
                        Response._display_dict(attribute, value, max_value_length, max_collection_length)
                    )
                    appended = True
                else:
                    value = {
                        Response._truncate_value(k_i, max_value_length): Response._truncate_value(v_i, max_value_length)
                        for k_i, v_i in value.items()
                    }
            else:
                value = Response._truncate_value(value, max_value_length)

            # appended == False <==> either the value was not a collection or it did not exceed the max length
            if not appended:
                message_parts.append(f"{attribute}={value!r}")

        message = ", ".join(message_parts)

        return f"{class_name}({message})"

    @staticmethod
    @no_type_check
    def _display_list(attribute: object, list_: list[bytes], max_value_length: int, max_collection_length: int) -> str:
        list_ = list_[:max_collection_length]
        return f"{attribute}=[{', '.join(str(Response._truncate_value(list_i, max_value_length)) for list_i in list_)}, ...]"  # noqa: E501

    @staticmethod
    @no_type_check
    def _display_set(attribute: object, set_: set[bytes], max_value_length: int, max_collection_length: int) -> str:
        set_ = list(set_)[:max_collection_length]
        return f"{attribute}={{{', '.join(str(Response._truncate_value(set_i, max_value_length)) for set_i in set_)}, ...}}"  # noqa: E501

    @staticmethod
    @no_type_check
    def _display_dict(
        attribute: object, dict_: dict[bytes, bytes], max_value_length: int, max_collection_length: int
    ) -> str:
        dict_ = dict(list(dict_.items())[:max_collection_length])
        return f"{attribute}={{{', '.join(str(Response._truncate_value(k_i, max_value_length)) + ': ' + str(Response._truncate_value(v_i, max_value_length)) for k_i, v_i in dict_.items())}, ...}}"  # noqa: E501

    @no_type_check
    def __repr__(self) -> str:
        return self._render_as_str(max_value_length=512, max_collection_length=1024)

    @no_type_check
    def __str__(self) -> str:
        return self._render_as_str()

    @staticmethod
    @no_type_check
    def _truncate_value(value: Any, max_length: int = 32) -> Any:
        if type(value) == bytes:
            if len(value) < max_length:
                return value

            return value[:max_length] + b"..."
        elif type(value) == str:
            if len(value) < max_length:
                return value

            return value[:max_length] + "..."
        else:
            return value

    @no_type_check
    def __eq__(self, other: object) -> bool:
        if other is None:
            return False

        if type(self) != type(other):
            return False

        return vars(self) == vars(other)

    @no_type_check
    def __hash__(self) -> int:
        state = [type(self)]
        state.extend(sorted(vars(self).items(), key=lambda x: x[0]))
        return hash(tuple(state))


class CacheResponse(Response):
    ...


class ControlResponse(Response):
    ...



from .control.cache.create import CreateCache, CreateCacheResponse
from .control.cache.delete import DeleteCache, DeleteCacheResponse
from .control.cache.list import ListCaches, ListCachesResponse
from .control.signing_key.create import CreateSigningKey, CreateSigningKeyResponse
from .control.signing_key.list import (
    ListSigningKeys,
    ListSigningKeysResponse,
    SigningKey,
)
from .control.signing_key.revoke import RevokeSigningKey, RevokeSigningKeyResponse
from .data.dictionary.fetch import CacheDictionaryFetch, CacheDictionaryFetchResponse
from .data.dictionary.get_field import (
    CacheDictionaryGetField,
    CacheDictionaryGetFieldResponse,
)
from .data.dictionary.get_fields import (
    CacheDictionaryGetFields,
    CacheDictionaryGetFieldsResponse,
)
from .data.dictionary.increment import (
    CacheDictionaryIncrement,
    CacheDictionaryIncrementResponse,
)
from .data.dictionary.remove_field import (
    CacheDictionaryRemoveField,
    CacheDictionaryRemoveFieldResponse,
)
from .data.dictionary.remove_fields import (
    CacheDictionaryRemoveFields,
    CacheDictionaryRemoveFieldsResponse,
)
from .data.dictionary.set_field import (
    CacheDictionarySetField,
    CacheDictionarySetFieldResponse,
)
from .data.dictionary.set_fields import (
    CacheDictionarySetFields,
    CacheDictionarySetFieldsResponse,
)
from .data.list.concatenate_back import (
    CacheListConcatenateBack,
    CacheListConcatenateBackResponse,
)
from .data.list.concatenate_front import (
    CacheListConcatenateFront,
    CacheListConcatenateFrontResponse,
)
from .data.list.fetch import CacheListFetch, CacheListFetchResponse
from .data.list.length import CacheListLength, CacheListLengthResponse
from .data.list.pop_back import CacheListPopBack, CacheListPopBackResponse
from .data.list.pop_front import CacheListPopFront, CacheListPopFrontResponse
from .data.list.push_back import CacheListPushBack, CacheListPushBackResponse
from .data.list.push_front import CacheListPushFront, CacheListPushFrontResponse
from .data.list.remove_value import CacheListRemoveValue, CacheListRemoveValueResponse
from .data.scalar.delete import CacheDelete, CacheDeleteResponse
from .data.scalar.get import CacheGet, CacheGetResponse
from .data.scalar.increment import CacheIncrement, CacheIncrementResponse
from .data.scalar.set import CacheSet, CacheSetResponse
from .data.scalar.set_if_not_exists import (
    CacheSetIfNotExists,
    CacheSetIfNotExistsResponse,
)
from .data.set.add_element import CacheSetAddElement, CacheSetAddElementResponse
from .data.set.add_elements import CacheSetAddElements, CacheSetAddElementsResponse
from .data.set.fetch import CacheSetFetch, CacheSetFetchResponse
from .data.set.remove_element import (
    CacheSetRemoveElement,
    CacheSetRemoveElementResponse,
)
from .data.set.remove_elements import (
    CacheSetRemoveElements,
    CacheSetRemoveElementsResponse,
)
from .response import CacheResponse, ControlResponse

__all__ = [
    "CreateCache",
    "CreateCacheResponse",
    "DeleteCache",
    "DeleteCacheResponse",
    "ListCaches",
    "ListCachesResponse",
    "CreateSigningKey",
    "CreateSigningKeyResponse",
    "ListSigningKeys",
    "ListSigningKeysResponse",
    "SigningKey",
    "RevokeSigningKey",
    "RevokeSigningKeyResponse",
    "CacheDictionaryFetch",
    "CacheDictionaryFetchResponse",
    "CacheDictionaryGetField",
    "CacheDictionaryGetFieldResponse",
    "CacheDictionaryGetFields",
    "CacheDictionaryGetFieldsResponse",
    "CacheDictionaryIncrement",
    "CacheDictionaryIncrementResponse",
    "CacheDictionaryRemoveField",
    "CacheDictionaryRemoveFieldResponse",
    "CacheDictionaryRemoveFields",
    "CacheDictionaryRemoveFieldsResponse",
    "CacheDictionarySetField",
    "CacheDictionarySetFieldResponse",
    "CacheDictionarySetFields",
    "CacheDictionarySetFieldsResponse",
    "CacheListConcatenateBack",
    "CacheListConcatenateBackResponse",
    "CacheListConcatenateFront",
    "CacheListConcatenateFrontResponse",
    "CacheListFetch",
    "CacheListFetchResponse",
    "CacheListLength",
    "CacheListLengthResponse",
    "CacheListPopBack",
    "CacheListPopBackResponse",
    "CacheListPopFront",
    "CacheListPopFrontResponse",
    "CacheListPushBack",
    "CacheListPushBackResponse",
    "CacheListPushFront",
    "CacheListPushFrontResponse",
    "CacheListRemoveValue",
    "CacheListRemoveValueResponse",
    "CacheDelete",
    "CacheDeleteResponse",
    "CacheGet",
    "CacheGetResponse",
    "CacheIncrement",
    "CacheIncrementResponse",
    "CacheSet",
    "CacheSetResponse",
    "CacheSetIfNotExists",
    "CacheSetIfNotExistsResponse",
    "CacheSetAddElement",
    "CacheSetAddElementResponse",
    "CacheSetAddElements",
    "CacheSetAddElementsResponse",
    "CacheSetFetch",
    "CacheSetFetchResponse",
    "CacheSetRemoveElement",
    "CacheSetRemoveElementResponse",
    "CacheSetRemoveElements",
    "CacheSetRemoveElementsResponse",
    "CacheResponse",
    "ControlResponse",
]



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Dict, Optional

from . import momento_endpoint_resolver


@dataclass
class CredentialProvider:
    """Information the CacheClient needs to connect to and authenticate with the Momento service."""

    auth_token: str
    control_endpoint: str
    cache_endpoint: str

    @staticmethod
    def from_environment_variable(
        env_var_name: str,
        control_endpoint: Optional[str] = None,
        cache_endpoint: Optional[str] = None,
    ) -> CredentialProvider:
        """Reads and parses a Momento auth token stored as an environment variable.

        Args:
            env_var_name (str): Name of the environment variable from which the auth token will be read
            control_endpoint (Optional[str], optional): Optionally overrides the default control endpoint.
            Defaults to None.
            cache_endpoint (Optional[str], optional): Optionally overrides the default control endpoint.
            Defaults to None.

        Raises:
            RuntimeError: if the environment variable is missing

        Returns:
            CredentialProvider
        """
        auth_token = os.getenv(env_var_name)
        if not auth_token:
            raise RuntimeError(f"Missing required environment variable {env_var_name}")
        return CredentialProvider.from_string(auth_token, control_endpoint, cache_endpoint)

    @staticmethod
    def from_string(
        auth_token: str,
        control_endpoint: Optional[str] = None,
        cache_endpoint: Optional[str] = None,
    ) -> CredentialProvider:
        """Reads and parses a Momento auth token.

        Args:
            auth_token (str): the Momento auth token
            control_endpoint (Optional[str], optional): Optionally overrides the default control endpoint.
            Defaults to None.
            cache_endpoint (Optional[str], optional): Optionally overrides the default control endpoint.
            Defaults to None.

        Returns:
            CredentialProvider
        """
        endpoints = momento_endpoint_resolver.resolve(auth_token)
        control_endpoint = control_endpoint or endpoints.control_endpoint
        cache_endpoint = cache_endpoint or endpoints.cache_endpoint
        return CredentialProvider(auth_token, control_endpoint, cache_endpoint)

    def __repr__(self) -> str:
        attributes: Dict[str, str] = vars(self)
        attributes["auth_token"] = self._obscure(attributes["auth_token"])
        message = ", ".join(f"{k}={v!r}" for k, v in attributes.items())
        return f"{self.__class__.__name__}({message})"

    def _obscure(self, value: str) -> str:
        return f"{value[:10]}...{value[-10:]}"



from abc import ABC, abstractmethod
from typing import Optional

from .retryable_props import RetryableProps


class RetryStrategy(ABC):
    @abstractmethod
    def determine_when_to_retry(self, props: RetryableProps) -> Optional[float]:
        pass



from __future__ import annotations

from typing import Optional, Tuple, Type

import grpc
from grpc.aio import Metadata

from momento.errors import (
    AlreadyExistsException,
    AuthenticationException,
    BadRequestException,
    CancelledException,
    FailedPreconditionException,
    InternalServerException,
    InvalidArgumentException,
    LimitExceededException,
    MomentoErrorTransportDetails,
    MomentoGrpcErrorDetails,
    NotFoundException,
    PermissionDeniedException,
    SdkException,
    ServerUnavailableException,
    TimeoutException,
    UnknownException,
    UnknownServiceException,
)

grpc_status_to_exception: dict[grpc.StatusCode, Type[SdkException]] = {
    grpc.StatusCode.INVALID_ARGUMENT: InvalidArgumentException,
    grpc.StatusCode.INTERNAL: InternalServerException,
    grpc.StatusCode.OUT_OF_RANGE: BadRequestException,
    grpc.StatusCode.UNIMPLEMENTED: BadRequestException,
    grpc.StatusCode.FAILED_PRECONDITION: FailedPreconditionException,
    grpc.StatusCode.PERMISSION_DENIED: PermissionDeniedException,
    grpc.StatusCode.UNAUTHENTICATED: AuthenticationException,
    grpc.StatusCode.RESOURCE_EXHAUSTED: LimitExceededException,
    grpc.StatusCode.NOT_FOUND: NotFoundException,
    grpc.StatusCode.ALREADY_EXISTS: AlreadyExistsException,
    grpc.StatusCode.DEADLINE_EXCEEDED: TimeoutException,
    grpc.StatusCode.CANCELLED: CancelledException,
    grpc.StatusCode.UNAVAILABLE: ServerUnavailableException,
    grpc.StatusCode.UNKNOWN: UnknownServiceException,
    grpc.StatusCode.ABORTED: InternalServerException,
    grpc.StatusCode.DATA_LOSS: InternalServerException,
}


INTERNAL_SERVER_ERROR_MESSAGE = "Unexpected exception occurred while trying to fulfill the request."
SDK_ERROR_MESSAGE = "SDK Failed to process the request."


def convert_error(
    exception: Exception, transport_metadata: Optional[Metadata | list[Tuple[str, str]]] = None
) -> SdkException:
    """Convert a low-level exception raised by gRPC to a Momento `SdkException`.

    Note about the metadata type: Metadata in the asynchronous gRPC client is of
    type Metadata, while metadata in the synchronous client is a list of tuples.
    The types are isomorphic.

    Args:
        exception (Exception): Low-level (transport, server-side, validation) exception
        transport_metadata (Optional[TMetadata], optional): Transport metadata to enrich the new exception.
        Defaults to None.

    Returns:
        SdkException: High-level Momento exception object
    """
    if isinstance(exception, SdkException):
        return exception

    # Normalize synchronous client metadata to `Metadata`
    if isinstance(transport_metadata, list):
        transport_metadata = _synchronous_metadata_to_metadata(transport_metadata)

    if isinstance(exception, grpc.RpcError):
        status_code: grpc.StatusCode = exception.code()
        details = exception.details()
        transport_details = MomentoErrorTransportDetails(
            MomentoGrpcErrorDetails(status_code, details, transport_metadata)
        )

        if status_code in grpc_status_to_exception:
            concrete_exception_type = grpc_status_to_exception[status_code]
            return concrete_exception_type(details, transport_details)  # type: ignore
        else:
            # TODO exception chaining from .NET redundant here?
            return InternalServerException(INTERNAL_SERVER_ERROR_MESSAGE, transport_details)

    # TODO exception chaining from .NET redundant here?
    return UnknownException(SDK_ERROR_MESSAGE, None)


def _synchronous_metadata_to_metadata(metadata: list[Tuple[str, str]]) -> Metadata:
    """Represent synchronous client metadata as a `Metadata` object.

    Args:
        metadata (List[Tuple[str, str]]): The synchronous client metadata.

    Returns:
        Metadata: async client metadata representation
    """
    new_metadata = Metadata()
    for key, value in metadata:
        new_metadata.add(key, value)
    return new_metadata



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheDictionaryFetchResponse(CacheResponse):
    """Response type for a `dictionary_fetch` request.

    Its subtypes are:
    - `CacheDictionaryFetch.Hit`
    - `CacheDictionaryFetch.Miss`
    - `CacheDictionaryFetch.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheDictionaryFetch(ABC):
    """Groups all `CacheDictionaryFetchResponse` derived types under a common namespace."""

    @dataclass
    class Hit(CacheDictionaryFetchResponse):
        """Indicates the dictionary exists and its items were fetched."""

        value_dictionary_bytes_bytes: dict[bytes, bytes]
        """The items for the fetched dictionary, as a mapping from bytes to bytes.

        Returns:
            dict[bytes, bytes]
        """

        @property
        def value_dictionary_string_bytes(self) -> dict[str, bytes]:
            """The items for the fetched dictionary, as a mapping from utf-8 encoded strings to bytes.

            Returns:
                dict[str, bytes]
            """
            return {k.decode("utf-8"): v for k, v in self.value_dictionary_bytes_bytes.items()}

        @property
        def value_dictionary_string_string(self) -> dict[str, str]:
            """The items for the fetched dictionary, as a mapping from utf-8 encoded strings to utf-8 encoded strings.

            Returns:
                dict[str, str]
            """
            return {k.decode("utf-8"): v.decode("utf-8") for k, v in self.value_dictionary_bytes_bytes.items()}

    class Miss(CacheDictionaryFetchResponse):
        """Indicates the dictionary does not exist."""

    class Error(CacheDictionaryFetchResponse, ErrorResponseMixin):
        """Indicates an error occured in the request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



import logging

from momento import logs

from .auth import CredentialProvider
from .cache_client import CacheClient
from .cache_client_async import CacheClientAsync
from .config import Configurations

logging.getLogger("momentosdk").addHandler(logging.NullHandler())
logs.initialize_momento_logging()

__all__ = ["CredentialProvider", "Configurations", "CacheClient", "CacheClientAsync"]



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import timedelta

from momento.internal._utilities import _validate_request_timeout

from .grpc_configuration import GrpcConfiguration


class TransportStrategy(ABC):
    """Configures the network options for communicating with the Momento service."""

    @abstractmethod
    def get_grpc_configuration(self) -> GrpcConfiguration:
        """Access the gRPC configuration.

        Returns:
            GrpcConfiguration: low-level gRPC settings for the Momento client's communication.
        """
        pass

    @abstractmethod
    def with_grpc_configuration(self, grpc_configuration: GrpcConfiguration) -> TransportStrategy:
        """Copy constructor for overriding the gRPC configuration.

        Args:
            grpc_configuration (GrpcConfiguration): the new gRPC configuration.

        Returns:
            TransportStrategy: a new TransportStrategy with the specified gRPC config.
        """
        pass

    @abstractmethod
    def with_client_timeout(self, client_timeout: timedelta) -> TransportStrategy:
        """Copies the TransportStrategy and updates the copy's client-side timeout.

        Args:
            client_timeout (timedelta): the new client-side timeout.

        Returns:
            TransportStrategy: the new TransportStrategy.
        """
        pass


class StaticGrpcConfiguration(GrpcConfiguration):
    def __init__(self, deadline: timedelta):
        self._deadline = deadline

    def get_deadline(self) -> timedelta:
        return self._deadline

    def with_deadline(self, deadline: timedelta) -> GrpcConfiguration:
        _validate_request_timeout(deadline)
        return StaticGrpcConfiguration(deadline)


class StaticTransportStrategy(TransportStrategy):
    def __init__(self, grpc_configuration: GrpcConfiguration):
        self._grpc_configuration = grpc_configuration

    def get_grpc_configuration(self) -> GrpcConfiguration:
        return self._grpc_configuration

    def with_grpc_configuration(self, grpc_configuration: GrpcConfiguration) -> TransportStrategy:
        return StaticTransportStrategy(grpc_configuration)

    def with_client_timeout(self, client_timeout: timedelta) -> TransportStrategy:
        return StaticTransportStrategy(self._grpc_configuration.with_deadline(client_timeout))



from __future__ import annotations

from datetime import timedelta
from typing import Optional

from momento_wire_types.cacheclient_pb2 import (
    Hit,
    Miss,
    _DeleteRequest,
    _DictionaryDeleteRequest,
    _DictionaryFetchRequest,
    _DictionaryFieldValuePair,
    _DictionaryGetRequest,
    _DictionaryIncrementRequest,
    _DictionarySetRequest,
    _GetRequest,
    _IncrementRequest,
    _ListConcatenateBackRequest,
    _ListConcatenateFrontRequest,
    _ListFetchRequest,
    _ListLengthRequest,
    _ListPopBackRequest,
    _ListPopFrontRequest,
    _ListPushBackRequest,
    _ListPushFrontRequest,
    _ListRemoveRequest,
    _SetDifferenceRequest,
    _SetFetchRequest,
    _SetIfNotExistsRequest,
    _SetRequest,
    _SetUnionRequest,
)
from momento_wire_types.cacheclient_pb2_grpc import ScsStub

from momento import logs
from momento.auth import CredentialProvider
from momento.config import Configuration
from momento.errors import UnknownException, convert_error
from momento.internal._utilities import (
    _as_bytes,
    _gen_dictionary_fields_as_bytes,
    _gen_dictionary_items_as_bytes,
    _gen_list_as_bytes,
    _gen_set_input_as_bytes,
    _validate_cache_name,
    _validate_dictionary_name,
    _validate_list_name,
    _validate_set_name,
    _validate_ttl,
)
from momento.internal.synchronous._scs_grpc_manager import _DataGrpcManager
from momento.internal.synchronous._utilities import make_metadata
from momento.requests import CollectionTtl
from momento.responses import (
    CacheDelete,
    CacheDeleteResponse,
    CacheDictionaryFetch,
    CacheDictionaryFetchResponse,
    CacheDictionaryGetField,
    CacheDictionaryGetFieldResponse,
    CacheDictionaryGetFields,
    CacheDictionaryGetFieldsResponse,
    CacheDictionaryIncrement,
    CacheDictionaryIncrementResponse,
    CacheDictionaryRemoveFields,
    CacheDictionaryRemoveFieldsResponse,
    CacheDictionarySetFields,
    CacheDictionarySetFieldsResponse,
    CacheGet,
    CacheGetResponse,
    CacheIncrement,
    CacheIncrementResponse,
    CacheListConcatenateBack,
    CacheListConcatenateBackResponse,
    CacheListConcatenateFront,
    CacheListConcatenateFrontResponse,
    CacheListFetch,
    CacheListFetchResponse,
    CacheListLength,
    CacheListLengthResponse,
    CacheListPopBack,
    CacheListPopBackResponse,
    CacheListPopFront,
    CacheListPopFrontResponse,
    CacheListPushBack,
    CacheListPushBackResponse,
    CacheListPushFront,
    CacheListPushFrontResponse,
    CacheListRemoveValue,
    CacheListRemoveValueResponse,
    CacheSet,
    CacheSetAddElements,
    CacheSetAddElementsResponse,
    CacheSetFetch,
    CacheSetFetchResponse,
    CacheSetIfNotExists,
    CacheSetIfNotExistsResponse,
    CacheSetRemoveElements,
    CacheSetRemoveElementsResponse,
    CacheSetResponse,
)
from momento.typing import (
    TCacheName,
    TDictionaryField,
    TDictionaryFields,
    TDictionaryItems,
    TDictionaryName,
    TListName,
    TListValue,
    TListValuesInput,
    TScalarKey,
    TScalarValue,
    TSetElementsInput,
    TSetName,
)


class _ScsDataClient:
    """Internal data client."""

    __UNSUPPORTED_LIST_NAME_TYPE_MSG = "Unsupported type for list_name: "
    __UNSUPPORTED_LIST_VALUE_TYPE_MSG = "Unsupported type for value: "
    __UNSUPPORTED_LIST_VALUES_TYPE_MSG = "Unsupported type for values: "
    __UNSUPPORTED_SET_NAME_TYPE_MSG = "Unsupported type for set_name: "
    __UNSUPPORTED_SET_ELEMENTS_TYPE_MSG = "Unsupported tyoe for elements: "

    __UNSUPPORTED_DICTIONARY_NAME_TYPE_MSG = "Unsupported type for dictionary_name: "
    __UNSUPPORTED_DICTIONARY_FIELD_TYPE_MSG = "Unsupported type for field: "
    __UNSUPPORTED_DICTIONARY_FIELDS_TYPE_MSG = "Unsupported type for fields: "
    __UNSUPPORTED_DICTIONARY_ITEMS_TYPE_MSG = "Unsupported type for items: "

    def __init__(self, configuration: Configuration, credential_provider: CredentialProvider, default_ttl: timedelta):
        endpoint = credential_provider.cache_endpoint
        self._logger = logs.logger
        self._logger.debug("Simple cache data client instantiated with endpoint: %s", endpoint)
        self._endpoint = endpoint

        default_deadline: timedelta = configuration.get_transport_strategy().get_grpc_configuration().get_deadline()
        self._default_deadline_seconds = int(default_deadline.total_seconds())

        self._grpc_manager = _DataGrpcManager(configuration, credential_provider)

        _validate_ttl(default_ttl)
        self._default_ttl = default_ttl

    @property
    def endpoint(self) -> str:
        return self._endpoint

    def increment(
        self, cache_name: TCacheName, key: TScalarKey, amount: int = 1, ttl: Optional[timedelta] = None
    ) -> CacheIncrementResponse:
        try:
            self._log_issuing_request("Increment", {"key": str(key), "amount": str(amount)})
            _validate_cache_name(cache_name)
            item_ttl = self._default_ttl if ttl is None else ttl
            _validate_ttl(item_ttl)

            request = _IncrementRequest()
            request.cache_key = _as_bytes(key, "Unsupported type for key: ")
            request.increment_by = amount
            request.ttl_milliseconds = int(item_ttl.total_seconds() * 1000)

            response = self._build_stub().Increment(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("Increment", {"key": str(key), "amount": str(amount)})
            return CacheIncrement.Success(response.value)
        except Exception as e:
            self._log_request_error("increment", e)
            return CacheIncrement.Error(convert_error(e))

    def set(
        self,
        cache_name: str,
        key: TScalarKey,
        value: TScalarValue,
        ttl: Optional[timedelta],
    ) -> CacheSetResponse:
        try:
            self._log_issuing_request("Set", {"key": str(key)})
            _validate_cache_name(cache_name)
            item_ttl = self._default_ttl if ttl is None else ttl
            _validate_ttl(item_ttl)
            request = _SetRequest()
            request.cache_key = _as_bytes(key, "Unsupported type for key: ")
            request.cache_body = _as_bytes(value, "Unsupported type for value: ")
            request.ttl_milliseconds = int(item_ttl.total_seconds() * 1000)

            self._build_stub().Set(request, metadata=make_metadata(cache_name), timeout=self._default_deadline_seconds)

            self._log_received_response("Set", {"key": str(key)})
            return CacheSet.Success()
        except Exception as e:
            self._log_request_error("set", e)
            return CacheSet.Error(convert_error(e))

    def set_if_not_exists(
        self, cache_name: TCacheName, key: TScalarKey, value: TScalarValue, ttl: Optional[timedelta]
    ) -> CacheSetIfNotExistsResponse:
        try:
            self._log_issuing_request("SetIfNotExists", {"key": str(key)})

            _validate_cache_name(cache_name)
            item_ttl = self._default_ttl if ttl is None else ttl
            _validate_ttl(item_ttl)
            request = _SetIfNotExistsRequest()
            request.cache_key = _as_bytes(key, "Unsupported type for key: ")
            request.cache_body = _as_bytes(value, "Unsupported type for value: ")
            request.ttl_milliseconds = int(item_ttl.total_seconds() * 1000)

            response = self._build_stub().SetIfNotExists(
                request, metadata=make_metadata(cache_name), timeout=self._default_deadline_seconds
            )

            self._log_received_response("SetIfNotExists", {"key": str(key)})

            result = response.WhichOneof("result")
            if result == "stored":
                return CacheSetIfNotExists.Stored()
            elif result == "not_stored":
                return CacheSetIfNotExists.NotStored()
            else:
                raise UnknownException("SetIfNotExists responded with an unknown result")
        except Exception as e:
            self._log_request_error("set_if_not_exists", e)
            return CacheSetIfNotExists.Error(convert_error(e))

    def get(self, cache_name: str, key: TScalarKey) -> CacheGetResponse:
        try:
            self._log_issuing_request("Get", {"key": str(key)})

            _validate_cache_name(cache_name)
            request = _GetRequest()
            request.cache_key = _as_bytes(key, "Unsupported type for key: ")

            response = self._build_stub().Get(
                request, metadata=make_metadata(cache_name), timeout=self._default_deadline_seconds
            )

            self._log_received_response("Get", {"key": str(key)})

            if response.result == Hit:
                return CacheGet.Hit(response.cache_body)
            elif response.result == Miss:
                return CacheGet.Miss()
            else:
                raise UnknownException("Get responded with an unknown result")
        except Exception as e:
            self._log_request_error("set_if_not_exists", e)
            return CacheGet.Error(convert_error(e))

    def delete(self, cache_name: str, key: TScalarKey) -> CacheDeleteResponse:
        try:
            self._log_issuing_request("Delete", {"key": str(key)})
            _validate_cache_name(cache_name)
            request = _DeleteRequest()
            request.cache_key = _as_bytes(key, "Unsupported type for key: ")

            self._build_stub().Delete(
                request, metadata=make_metadata(cache_name), timeout=self._default_deadline_seconds
            )

            self._log_received_response("Delete", {"key": str(key)})
            return CacheDelete.Success()
        except Exception as e:
            self._log_request_error("set", e)
            return CacheDelete.Error(convert_error(e))

    # DICTIONARY COLLECTION METHODS
    def dictionary_get_fields(
        self,
        cache_name: TCacheName,
        dictionary_name: TDictionaryName,
        fields: TDictionaryFields,
    ) -> CacheDictionaryGetFieldsResponse:
        try:
            self._log_issuing_request("DictionaryGet", {"dictionary_name": dictionary_name})
            _validate_cache_name(cache_name)
            _validate_dictionary_name(dictionary_name)

            request = _DictionaryGetRequest()
            request.dictionary_name = _as_bytes(dictionary_name, self.__UNSUPPORTED_DICTIONARY_NAME_TYPE_MSG)
            bytes_fields = list(_gen_dictionary_fields_as_bytes(fields, self.__UNSUPPORTED_DICTIONARY_FIELDS_TYPE_MSG))

            request.fields.extend(bytes_fields)

            response = self._build_stub().DictionaryGet(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("DictionaryGet", {"dictionary_name": dictionary_name})

            type = response.WhichOneof("dictionary")
            if type == "found":
                get_responses: list[CacheDictionaryGetFieldResponse] = []
                for field, get_response in zip(bytes_fields, response.found.items):
                    if get_response.result == Miss:
                        get_responses.append(CacheDictionaryGetField.Miss())
                    else:
                        get_responses.append(CacheDictionaryGetField.Hit(get_response.cache_body, field))
                return CacheDictionaryGetFields.Hit(get_responses)
            elif type == "missing":
                return CacheDictionaryGetFields.Miss()
            else:
                raise UnknownException("Unknown dictionary field")
        except Exception as e:
            self._log_request_error("dictionary_get_fields", e)
            return CacheDictionaryGetFields.Error(convert_error(e))

    def dictionary_fetch(
        self, cache_name: TCacheName, dictionary_name: TDictionaryName
    ) -> CacheDictionaryFetchResponse:
        try:
            self._log_issuing_request("DictionaryFetch", {"dictionary_name": dictionary_name})
            _validate_cache_name(cache_name)
            _validate_dictionary_name(dictionary_name)
            request = _DictionaryFetchRequest()
            request.dictionary_name = _as_bytes(dictionary_name, self.__UNSUPPORTED_DICTIONARY_NAME_TYPE_MSG)
            response = self._build_stub().DictionaryFetch(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("DictionaryFetch", {"dictionary_name": dictionary_name})

            type = response.WhichOneof("dictionary")
            if type == "missing":
                return CacheDictionaryFetch.Miss()
            elif type == "found":
                return CacheDictionaryFetch.Hit({item.field: item.value for item in response.found.items})
            else:
                raise UnknownException("Unknown dictionary field")
        except Exception as e:
            self._log_request_error("dictionary_fetch", e)
            return CacheDictionaryFetch.Error(convert_error(e))

    def dictionary_increment(
        self,
        cache_name: TCacheName,
        dictionary_name: TDictionaryName,
        field: TDictionaryField,
        amount: int = 1,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheDictionaryIncrementResponse:
        try:
            self._log_issuing_request("DictionaryIncrement", {"dictionary_name": dictionary_name})
            _validate_cache_name(cache_name)
            _validate_dictionary_name(dictionary_name)

            request = _DictionaryIncrementRequest()
            request.dictionary_name = _as_bytes(dictionary_name, self.__UNSUPPORTED_DICTIONARY_NAME_TYPE_MSG)
            request.field = _as_bytes(field, self.__UNSUPPORTED_DICTIONARY_FIELD_TYPE_MSG)
            request.amount = amount
            request.ttl_milliseconds = self._collection_ttl_or_default_milliseconds(ttl)
            request.refresh_ttl = ttl.refresh_ttl

            response = self._build_stub().DictionaryIncrement(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("DictionaryIncrement", {"dictionary_name": dictionary_name})
            return CacheDictionaryIncrement.Success(response.value)
        except Exception as e:
            self._log_request_error("dictionary_increment", e)
            return CacheDictionaryIncrement.Error(convert_error(e))

    def dictionary_remove_fields(
        self,
        cache_name: TCacheName,
        dictionary_name: TDictionaryName,
        fields: TDictionaryFields,
    ) -> CacheDictionaryRemoveFieldsResponse:
        try:
            self._log_issuing_request("DictionaryDelete", {"dictionary_name": dictionary_name})
            _validate_cache_name(cache_name)
            _validate_dictionary_name(dictionary_name)

            request = _DictionaryDeleteRequest()
            request.dictionary_name = _as_bytes(dictionary_name, self.__UNSUPPORTED_DICTIONARY_NAME_TYPE_MSG)
            request.some.fields.extend(
                _gen_dictionary_fields_as_bytes(fields, self.__UNSUPPORTED_DICTIONARY_FIELDS_TYPE_MSG)
            )

            self._build_stub().DictionaryDelete(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("DictionaryDelete", {"dictionary_name": dictionary_name})
            return CacheDictionaryRemoveFields.Success()
        except Exception as e:
            self._log_request_error("dictionary_remove_fields", e)
            return CacheDictionaryRemoveFields.Error(convert_error(e))

    def dictionary_set_fields(
        self,
        cache_name: TCacheName,
        dictionary_name: TDictionaryName,
        items: TDictionaryItems,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheDictionarySetFieldsResponse:
        try:
            self._log_issuing_request("DictionarySet", {"dictionary_name": dictionary_name})
            _validate_cache_name(cache_name)
            _validate_dictionary_name(dictionary_name)

            request = _DictionarySetRequest()
            request.dictionary_name = _as_bytes(dictionary_name, self.__UNSUPPORTED_DICTIONARY_NAME_TYPE_MSG)
            for field, value in _gen_dictionary_items_as_bytes(items, self.__UNSUPPORTED_DICTIONARY_ITEMS_TYPE_MSG):
                field_value_pair = _DictionaryFieldValuePair()
                field_value_pair.field = field
                field_value_pair.value = value
                request.items.append(field_value_pair)
            request.ttl_milliseconds = self._collection_ttl_or_default_milliseconds(ttl)
            request.refresh_ttl = ttl.refresh_ttl

            self._build_stub().DictionarySet(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("DictionarySet", {"dictionary_name": dictionary_name})
            return CacheDictionarySetFields.Success()
        except Exception as e:
            self._log_request_error("dictionary_set_fields", e)
            return CacheDictionarySetFields.Error(convert_error(e))

    # LIST COLLECTION METHODS
    def list_concatenate_back(
        self,
        cache_name: TCacheName,
        list_name: TListName,
        values: TListValuesInput,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_front_to_size: Optional[int] = None,
    ) -> CacheListConcatenateBackResponse:
        try:
            self._log_issuing_request("ListConcatenateBack", {})
            _validate_cache_name(cache_name)
            _validate_list_name(list_name)

            request = _ListConcatenateBackRequest()
            request.list_name = _as_bytes(list_name, self.__UNSUPPORTED_LIST_NAME_TYPE_MSG)
            request.values.extend(_gen_list_as_bytes(values, self.__UNSUPPORTED_LIST_VALUES_TYPE_MSG))
            request.ttl_milliseconds = self._collection_ttl_or_default_milliseconds(ttl)
            request.refresh_ttl = ttl.refresh_ttl
            if truncate_front_to_size is not None:
                request.truncate_front_to_size = truncate_front_to_size

            response = self._build_stub().ListConcatenateBack(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("ListConcatenateBack", {"list_name": str(request.list_name)})
            return CacheListConcatenateBack.Success(response.list_length)
        except Exception as e:
            self._log_request_error("list_concatenate_back", e)
            return CacheListConcatenateBack.Error(convert_error(e))

    def list_concatenate_front(
        self,
        cache_name: TCacheName,
        list_name: TListName,
        values: TListValuesInput,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_back_to_size: Optional[int] = None,
    ) -> CacheListConcatenateFrontResponse:
        try:
            self._log_issuing_request("ListConcatenateFront", {})
            _validate_cache_name(cache_name)
            _validate_list_name(list_name)

            request = _ListConcatenateFrontRequest()
            request.list_name = _as_bytes(list_name, self.__UNSUPPORTED_LIST_NAME_TYPE_MSG)
            request.values.extend(_gen_list_as_bytes(values, self.__UNSUPPORTED_LIST_VALUES_TYPE_MSG))
            request.ttl_milliseconds = self._collection_ttl_or_default_milliseconds(ttl)
            request.refresh_ttl = ttl.refresh_ttl
            if truncate_back_to_size is not None:
                request.truncate_back_to_size = truncate_back_to_size

            response = self._build_stub().ListConcatenateFront(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("ListConcatenateFront", {"list_name": str(request.list_name)})
            return CacheListConcatenateFront.Success(response.list_length)
        except Exception as e:
            self._log_request_error("list_concatenate_front", e)
            return CacheListConcatenateFront.Error(convert_error(e))

    def list_fetch(self, cache_name: TCacheName, list_name: TListName) -> CacheListFetchResponse:
        try:
            self._log_issuing_request("ListFetch", {"list_name": str(list_name)})
            _validate_cache_name(cache_name)
            _validate_list_name(list_name)
            request = _ListFetchRequest()
            request.list_name = _as_bytes(list_name, self.__UNSUPPORTED_LIST_NAME_TYPE_MSG)
            response = self._build_stub().ListFetch(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("ListFetch", {"list_name": str(request.list_name)})

            type = response.WhichOneof("list")
            if type == "missing":
                return CacheListFetch.Miss()
            elif type == "found":
                return CacheListFetch.Hit(list(response.found.values))
            else:
                raise UnknownException("Unknown list field")
        except Exception as e:
            self._log_request_error("list_fetch", e)
            return CacheListFetch.Error(convert_error(e))

    def list_length(self, cache_name: TCacheName, list_name: TListName) -> CacheListLengthResponse:
        try:
            self._log_issuing_request("ListLength", {"list_name": str(list_name)})
            _validate_cache_name(cache_name)
            _validate_list_name(list_name)
            request = _ListLengthRequest()
            request.list_name = _as_bytes(list_name, self.__UNSUPPORTED_LIST_NAME_TYPE_MSG)
            response = self._build_stub().ListLength(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("ListLength", {"list_name": str(request.list_name)})

            type = response.WhichOneof("list")
            if type == "missing":
                return CacheListLength.Miss()
            elif type == "found":
                return CacheListLength.Hit(response.found.length)
            else:
                raise UnknownException("Unknown list field")
        except Exception as e:
            self._log_request_error("list_length", e)
            return CacheListLength.Error(convert_error(e))

    def list_pop_back(self, cache_name: TCacheName, list_name: TListName) -> CacheListPopBackResponse:
        try:
            self._log_issuing_request("ListPopBack", {"list_name": str(list_name)})
            _validate_cache_name(cache_name)
            _validate_list_name(list_name)
            request = _ListPopBackRequest()
            request.list_name = _as_bytes(list_name, self.__UNSUPPORTED_LIST_NAME_TYPE_MSG)
            response = self._build_stub().ListPopBack(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("ListPopBack", {"list_name": str(request.list_name)})

            type = response.WhichOneof("list")
            if type == "missing":
                return CacheListPopBack.Miss()
            elif type == "found":
                return CacheListPopBack.Hit(response.found.back)
            else:
                raise UnknownException("Unknown list field")
        except Exception as e:
            self._log_request_error("list_pop_back", e)
            return CacheListPopBack.Error(convert_error(e))

    def list_pop_front(self, cache_name: TCacheName, list_name: TListName) -> CacheListPopFrontResponse:
        try:
            self._log_issuing_request("ListPopFront", {"list_name": str(list_name)})
            _validate_cache_name(cache_name)
            _validate_list_name(list_name)
            request = _ListPopFrontRequest()
            request.list_name = _as_bytes(list_name, self.__UNSUPPORTED_LIST_NAME_TYPE_MSG)
            response = self._build_stub().ListPopFront(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("ListPopFront", {"list_name": str(request.list_name)})

            type = response.WhichOneof("list")
            if type == "missing":
                return CacheListPopFront.Miss()
            elif type == "found":
                return CacheListPopFront.Hit(response.found.front)
            else:
                raise UnknownException("Unknown list field")
        except Exception as e:
            self._log_request_error("list_pop_front", e)
            return CacheListPopFront.Error(convert_error(e))

    def list_push_back(
        self,
        cache_name: TCacheName,
        list_name: TListName,
        value: TListValue,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_front_to_size: Optional[int] = None,
    ) -> CacheListPushBackResponse:
        try:
            self._log_issuing_request("ListPushBack", {})
            _validate_cache_name(cache_name)
            _validate_list_name(list_name)

            request = _ListPushBackRequest()
            request.list_name = _as_bytes(list_name, self.__UNSUPPORTED_LIST_NAME_TYPE_MSG)
            request.value = _as_bytes(value, self.__UNSUPPORTED_LIST_VALUE_TYPE_MSG)
            request.ttl_milliseconds = self._collection_ttl_or_default_milliseconds(ttl)
            request.refresh_ttl = ttl.refresh_ttl
            if truncate_front_to_size is not None:
                request.truncate_front_to_size = truncate_front_to_size

            response = self._build_stub().ListPushBack(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("ListPushBack", {"list_name": str(request.list_name)})
            return CacheListPushBack.Success(response.list_length)
        except Exception as e:
            self._log_request_error("list_push_back", e)
            return CacheListPushBack.Error(convert_error(e))

    def list_push_front(
        self,
        cache_name: TCacheName,
        list_name: TListName,
        value: TListValue,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_back_to_size: Optional[int] = None,
    ) -> CacheListPushFrontResponse:
        try:
            self._log_issuing_request("ListPushFront", {})
            _validate_cache_name(cache_name)
            _validate_list_name(list_name)

            request = _ListPushFrontRequest()
            request.list_name = _as_bytes(list_name, self.__UNSUPPORTED_LIST_NAME_TYPE_MSG)
            request.value = _as_bytes(value, self.__UNSUPPORTED_LIST_VALUE_TYPE_MSG)
            request.ttl_milliseconds = self._collection_ttl_or_default_milliseconds(ttl)
            request.refresh_ttl = ttl.refresh_ttl
            if truncate_back_to_size is not None:
                request.truncate_back_to_size = truncate_back_to_size

            response = self._build_stub().ListPushFront(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("ListPushFront", {"list_name": str(request.list_name)})
            return CacheListPushFront.Success(response.list_length)
        except Exception as e:
            self._log_request_error("list_push_front", e)
            return CacheListPushFront.Error(convert_error(e))

    def list_remove_value(
        self,
        cache_name: TCacheName,
        list_name: TListName,
        value: TListValue,
    ) -> CacheListRemoveValueResponse:
        try:
            self._log_issuing_request("ListRemoveValue", {})
            _validate_cache_name(cache_name)
            _validate_list_name(list_name)

            request = _ListRemoveRequest()
            request.list_name = _as_bytes(list_name, self.__UNSUPPORTED_LIST_NAME_TYPE_MSG)
            request.all_elements_with_value = _as_bytes(value, self.__UNSUPPORTED_LIST_VALUE_TYPE_MSG)

            self._build_stub().ListRemove(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("ListRemoveValue", {"list_name": str(request.list_name)})
            return CacheListRemoveValue.Success()
        except Exception as e:
            self._log_request_error("list_remove_value", e)
            return CacheListRemoveValue.Error(convert_error(e))

    # SET COLLECTION METHODS
    def set_add_elements(
        self,
        cache_name: TCacheName,
        set_name: TSetName,
        elements: TSetElementsInput,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheSetAddElementsResponse:
        try:
            self._log_issuing_request("SetAddElements", {})
            _validate_cache_name(cache_name)
            _validate_set_name(set_name)

            request = _SetUnionRequest()
            request.set_name = _as_bytes(set_name, self.__UNSUPPORTED_SET_NAME_TYPE_MSG)
            request.elements.extend(_gen_set_input_as_bytes(elements, self.__UNSUPPORTED_SET_ELEMENTS_TYPE_MSG))
            request.ttl_milliseconds = self._collection_ttl_or_default_milliseconds(ttl)
            request.refresh_ttl = ttl.refresh_ttl

            self._build_stub().SetUnion(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("SetAddElements", {"set_name": str(request.set_name)})
            return CacheSetAddElements.Success()
        except Exception as e:
            self._log_request_error("set_remove_elements", e)
            return CacheSetAddElements.Error(convert_error(e))

    def set_fetch(
        self,
        cache_name: TCacheName,
        set_name: TSetName,
    ) -> CacheSetFetchResponse:
        try:
            self._log_issuing_request("SetFetch", {"set_name": str(set_name)})
            _validate_cache_name(cache_name)
            _validate_set_name(set_name)

            request = _SetFetchRequest()
            request.set_name = _as_bytes(set_name, "Unsupported type for set_name: ")
            response = self._build_stub().SetFetch(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("SetFetch", {"set_name": str(request.set_name)})

            type = response.WhichOneof("set")
            if type == "missing":
                return CacheSetFetch.Miss()
            elif type == "found":
                return CacheSetFetch.Hit(set(response.found.elements))
            else:
                raise UnknownException(f"Unknown set field in response: {type}")
        except Exception as e:
            self._log_request_error("set_fetch", e)
            return CacheSetFetch.Error(convert_error(e))

    def set_remove_elements(
        self, cache_name: TCacheName, set_name: TSetName, elements: TSetElementsInput
    ) -> CacheSetRemoveElementsResponse:
        try:
            self._log_issuing_request("SetRemoveElements", {})
            _validate_cache_name(cache_name)
            _validate_set_name(set_name)

            request = _SetDifferenceRequest()
            request.set_name = _as_bytes(set_name, self.__UNSUPPORTED_SET_NAME_TYPE_MSG)
            request.subtrahend.set.elements.extend(
                _gen_set_input_as_bytes(elements, self.__UNSUPPORTED_SET_ELEMENTS_TYPE_MSG)
            )

            self._build_stub().SetDifference(
                request,
                metadata=make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._log_received_response("SetRemoveElements", {"set_name": str(request.set_name)})
            return CacheSetRemoveElements.Success()
        except Exception as e:
            self._log_request_error("set_remove_elements", e)
            return CacheSetRemoveElements.Error(convert_error(e))

    def _log_received_response(self, request_type: str, request_args: dict[str, str]) -> None:
        self._logger.log(logs.TRACE, f"Received a {request_type} response for {request_args}")

    def _log_issuing_request(self, request_type: str, request_args: dict[str, str]) -> None:
        self._logger.log(logs.TRACE, f"Issuing a {request_type} request with {request_args}")

    def _log_request_error(self, request_type: str, e: Exception) -> None:
        self._logger.warning(f"{request_type} failed with exception: {e}")

    def _collection_ttl_or_default_milliseconds(self, collection_ttl: CollectionTtl) -> int:
        return self._ttl_or_default_milliseconds(collection_ttl.ttl)

    def _ttl_or_default_milliseconds(self, ttl: Optional[timedelta]) -> int:
        which_ttl = self._default_ttl
        if ttl is not None:
            which_ttl = ttl

        return int(which_ttl.total_seconds() * 1000)

    def _build_stub(self) -> ScsStub:
        return self._grpc_manager.stub()

    def close(self) -> None:
        self._grpc_manager.close()



from __future__ import annotations

from typing import Callable

import grpc
from grpc.aio import ClientCallDetails, Metadata

from momento.errors import InvalidArgumentException


class Header:
    once_only_headers = ["agent"]

    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value


class AddHeaderClientInterceptor(grpc.aio.UnaryUnaryClientInterceptor):
    are_only_once_headers_sent = False

    def __init__(self, headers: list[Header]):
        self._headers_to_add_once: list[Header] = list(
            filter(lambda header: header.name in header.once_only_headers, headers)
        )
        self.headers_to_add_every_time = list(
            filter(lambda header: header.name not in header.once_only_headers, headers)
        )

    async def intercept_unary_unary(
        self,
        continuation: Callable[
            [grpc.aio._interceptor.ClientCallDetails, grpc.aio._typing.RequestType],
            grpc.aio._call.UnaryUnaryCall,
        ],
        client_call_details: grpc.aio._interceptor.ClientCallDetails,
        request: grpc.aio._typing.RequestType,
    ) -> grpc.aio._call.UnaryUnaryCall | grpc.aio._typing.ResponseType:

        new_client_call_details = sanitize_client_call_details(client_call_details)

        for header in self.headers_to_add_every_time:
            new_client_call_details.metadata.add(header.name, header.value)

        if not AddHeaderClientInterceptor.are_only_once_headers_sent:
            for header in self._headers_to_add_once:
                new_client_call_details.metadata.add(header.name, header.value)
                AddHeaderClientInterceptor.are_only_once_headers_sent = True

        return await continuation(new_client_call_details, request)


def sanitize_client_call_details(client_call_details: grpc.aio.ClientCallDetails) -> grpc.aio.ClientCallDetails:
    """Defensive function meant to handle inbound gRPC client request objects.

    Args:
        client_call_details: the original inbound client grpc request we are intercepting

    Returns: a new client_call_details object with metadata properly initialized to a `grpc.aio.Metadata` object
    """
    # Makes sure we can handle properly when we inject our own metadata onto request object.
    # This was mainly done as temporary fix after we observed ddtrace grpc client interceptor passing
    # client_call_details.metadata as a list instead of a grpc.aio.Metadata object.
    # See this ticket for follow-up actions to come back in and address this longer term:
    # https://github.com/momentohq/client-sdk-python/issues/149
    new_client_call_details = None
    # If no metadata set on passed in client call details then we are first to set, so we should just initialize
    if client_call_details.metadata is None:
        new_client_call_details = ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=Metadata(),
            credentials=client_call_details.credentials,
            wait_for_ready=client_call_details.wait_for_ready,
        )

    # This is block hit when ddtrace interceptor runs first and sets metadata as a list
    elif isinstance(client_call_details.metadata, list):
        existing_headers = client_call_details.metadata
        metadata = Metadata()
        # re-add all existing values to new metadata
        for md_key, md_value in existing_headers:
            metadata.add(md_key, md_value)
        new_client_call_details = ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=metadata,
            credentials=client_call_details.credentials,
            wait_for_ready=client_call_details.wait_for_ready,
        )
    elif isinstance(client_call_details.metadata, grpc.aio.Metadata):
        # If  proper grpc `grpc.aio.Metadata()` object is passed just use original object passed and pass back
        new_client_call_details = client_call_details
    else:
        # Else we raise exception for now since we don't know how to handle an unknown type
        raise InvalidArgumentException(
            "unexpected grpc client request metadata property passed to interceptor "
            "type=" + str(type(client_call_details.metadata))
        )

    return new_client_call_details



from abc import ABC, abstractmethod

from .retryable_props import RetryableProps


class EligibilityStrategy(ABC):
    @abstractmethod
    def is_eligible_for_retry(self, props: RetryableProps) -> bool:
        pass



from abc import ABC
from dataclasses import dataclass

from momento.errors import SdkException

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheIncrementResponse(CacheResponse):
    """Parent response type for a cache `increment` request.

    Its subtypes are:
    - `CacheIncrement.Success`
    - `CacheIncrement.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheIncrement(ABC):
    """Groups all `CacheIncrementResponse` derived types under a common namespace."""

    @dataclass
    class Success(CacheIncrementResponse):
        """Indicates the increment was successful."""

        value: int
        """The value of the field post-increment."""

    @dataclass
    class Error(CacheIncrementResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """

        _error: SdkException

        def __init__(self, _error: SdkException):
            self._error = _error



from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheDictionaryIncrementResponse(CacheResponse):
    """Parent response type for a cache `dictionary_increment` request.

    Its subtypes are:
    - `CacheDictionaryIncrement.Success`
    - `CacheDictionaryIncrement.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheDictionaryIncrement(ABC):
    """Groups all `CacheDictionaryIncrementResponse` derived types under a common namespace."""

    @dataclass
    class Success(CacheDictionaryIncrementResponse):
        """Indicates the request was successful."""

        value: int
        """The value of the field post-increment."""

    class Error(CacheDictionaryIncrementResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from abc import ABC

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheSetRemoveElementsResponse(CacheResponse):
    """Parent response type for a `set_remove_elements` request.

    Its subtypes are:
    - `CacheSetRemoveElements.Success`
    - `CacheSetRemoveElements.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheSetRemoveElements(ABC):
    """Groups all `CacheSetRemoveElementsResponse` derived types under a common namespace."""

    class Success(CacheSetRemoveElementsResponse):
        """Indicates the elements were removed."""

    class Error(CacheSetRemoveElementsResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from abc import ABC
from dataclasses import dataclass

from ...mixins import ErrorResponseMixin, ValueStringMixin
from ...response import CacheResponse


class CacheGetResponse(CacheResponse):
    """Parent response type for a cache `get` request.

    Its subtypes are:
    - `CacheGet.Hit`
    - `CacheGet.Miss`
    - `CacheGet.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheGet(ABC):
    """Groups all `CacheGetResponse` derived types under a common namespace."""

    @dataclass
    class Hit(CacheGetResponse, ValueStringMixin):
        """Contains the result of a cache hit."""

        value_bytes: bytes
        """The value returned from the cache for the specified key. Use the
        `value_string` property to access the value as a string."""

    class Miss(CacheGetResponse):
        """Contains the results of a cache miss."""

    class Error(CacheGetResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

import collections
from typing import Callable, TypeVar

import grpc

from momento.errors import InvalidArgumentException

RequestType = TypeVar("RequestType")
ResponseType = TypeVar("ResponseType")


class Header:
    once_only_headers = ["agent"]

    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value


class _ClientCallDetails(
    collections.namedtuple("_ClientCallDetails", ("method", "timeout", "metadata", "credentials")),
    grpc.ClientCallDetails,
):
    pass


class AddHeaderClientInterceptor(grpc.UnaryUnaryClientInterceptor):
    are_only_once_headers_sent = False

    @staticmethod
    def is_only_once_header(header: Header) -> bool:
        return header.name in header.once_only_headers

    @staticmethod
    def is_not_only_once_header(header: Header) -> bool:
        return header.name not in header.once_only_headers

    def __init__(self, headers: list[Header]):
        self._headers_to_add_once: list[Header] = list(filter(AddHeaderClientInterceptor.is_only_once_header, headers))
        self.headers_to_add_every_time = list(filter(AddHeaderClientInterceptor.is_not_only_once_header, headers))

    def intercept_unary_unary(
        self,
        continuation: Callable[[grpc.ClientCallDetails, RequestType], grpc.Call],
        client_call_details: grpc.ClientCallDetails,
        request: RequestType,
    ) -> grpc.Call:
        new_client_call_details = sanitize_client_call_details(client_call_details)

        for header in self.headers_to_add_every_time:
            new_client_call_details.metadata.append((header.name, header.value))

        if not AddHeaderClientInterceptor.are_only_once_headers_sent:
            for header in self._headers_to_add_once:
                new_client_call_details.metadata.append((header.name, header.value))
                AddHeaderClientInterceptor.are_only_once_headers_sent = True

        return continuation(new_client_call_details, request)


def sanitize_client_call_details(client_call_details: grpc.ClientCallDetails) -> grpc.ClientCallDetails:
    """Defensive function meant to handle inbound gRPC client request objects.

    Args:
        client_call_details: the original inbound client grpc request we are intercepting

    Returns: a new client_call_details object with metadata properly initialized to a `grpc.aio.Metadata` object
    """
    # Makes sure we can handle properly when we inject our own metadata onto request object.
    # This was mainly done as temporary fix after we observed ddtrace grpc client interceptor passing
    # client_call_details.metadata as a list instead of a grpc.aio.Metadata object.
    # See this ticket for follow-up actions to come back in and address this longer term:
    #    https://github.com/momentohq/client-sdk-python/issues/149
    # If no metadata set on passed in client call details then we are first to set, so we should just initialize
    if client_call_details.metadata is None:
        return _ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=[],
            credentials=client_call_details.credentials,
        )

    # This is block hit when ddtrace interceptor runs first and sets metadata as a list
    elif isinstance(client_call_details.metadata, list):
        return client_call_details
    else:
        # Else we raise exception for now since we don't know how to handle an unknown type
        raise InvalidArgumentException(
            "unexpected grpc client request metadata property passed to interceptor "
            "type=" + str(type(client_call_details.metadata))
        )



from __future__ import annotations

import grpc
import momento_wire_types.cacheclient_pb2_grpc as cache_client
import momento_wire_types.controlclient_pb2_grpc as control_client
import pkg_resources

from momento.auth import CredentialProvider
from momento.config import Configuration
from momento.retry import RetryStrategy

from ._add_header_client_interceptor import AddHeaderClientInterceptor, Header
from ._retry_interceptor import RetryInterceptor


class _ControlGrpcManager:
    """Internal gRPC control mananger."""

    version = pkg_resources.get_distribution("momento").version

    def __init__(self, configuration: Configuration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.aio.secure_channel(
            target=credential_provider.control_endpoint,
            credentials=grpc.ssl_channel_credentials(),
            interceptors=_interceptors(credential_provider.auth_token, configuration.get_retry_strategy()),
        )

    async def close(self) -> None:
        await self._secure_channel.close()

    def async_stub(self) -> control_client.ScsControlStub:
        return control_client.ScsControlStub(self._secure_channel)


class _DataGrpcManager:
    """Internal gRPC data mananger."""

    version = pkg_resources.get_distribution("momento").version

    def __init__(self, configuration: Configuration, credential_provider: CredentialProvider):
        self._secure_channel = grpc.aio.secure_channel(
            target=credential_provider.cache_endpoint,
            credentials=grpc.ssl_channel_credentials(),
            interceptors=_interceptors(credential_provider.auth_token, configuration.get_retry_strategy()),
            # Here is where you would pass override configuration to the underlying C gRPC layer.
            # However, I have tried several different tuning options here and did not see any
            # performance improvements, so sticking with the defaults for now.
            # For more info on the performance investigations:
            # https://github.com/momentohq/client-sdk-python/issues/120
            # For more info on available gRPC config options:
            # https://grpc.github.io/grpc/python/grpc.html
            # https://grpc.github.io/grpc/python/glossary.html#term-channel_arguments
            # https://github.com/grpc/grpc/blob/v1.46.x/include/grpc/impl/codegen/grpc_types.h#L140
            options=[
                # ('grpc.max_concurrent_streams', 1000),
                # ('grpc.use_local_subchannel_pool', 1),
                # (experimental.ChannelOptions.SingleThreadedUnaryStream, 1)
            ],
        )

    async def close(self) -> None:
        await self._secure_channel.close()

    def async_stub(self) -> cache_client.ScsStub:
        return cache_client.ScsStub(self._secure_channel)


def _interceptors(auth_token: str, retry_strategy: RetryStrategy) -> list[grpc.aio.ClientInterceptor]:
    headers = [
        Header("authorization", auth_token),
        Header("agent", f"python:{_ControlGrpcManager.version}"),
    ]
    return [
        AddHeaderClientInterceptor(headers),
        RetryInterceptor(retry_strategy),
    ]



from __future__ import annotations

from datetime import timedelta
from types import TracebackType
from typing import Iterable, Optional, Type

from momento import logs
from momento.auth import CredentialProvider
from momento.config import Configuration
from momento.errors import UnknownException
from momento.requests import CollectionTtl

try:
    from momento.internal._utilities import _validate_request_timeout
    from momento.internal.aio._scs_control_client import _ScsControlClient
    from momento.internal.aio._scs_data_client import _ScsDataClient
except ImportError as e:
    if e.name == "cygrpc":
        import sys

        print(
            "There is an issue on M1 macs between GRPC native packaging and Python wheel tags. "
            "See https://github.com/grpc/grpc/issues/28387",
            file=sys.stderr,
        )
        print("-".join("" for _ in range(99)), file=sys.stderr)
        print("    TO WORK AROUND:", file=sys.stderr)
        print("    * Install Rosetta 2", file=sys.stderr)
        print(
            "    * Install Python from python.org (you might need to do this if you're using an arm-only build)",
            file=sys.stderr,
        )
        print("    * re-run with:", file=sys.stderr)
        print("arch -x86_64 {} {}".format(sys.executable, *sys.argv), file=sys.stderr)
        print("-".join("" for _ in range(99)), file=sys.stderr)
    raise e

from momento.responses import (
    CacheDeleteResponse,
    CacheDictionaryFetchResponse,
    CacheDictionaryGetField,
    CacheDictionaryGetFieldResponse,
    CacheDictionaryGetFields,
    CacheDictionaryGetFieldsResponse,
    CacheDictionaryIncrementResponse,
    CacheDictionaryRemoveField,
    CacheDictionaryRemoveFieldResponse,
    CacheDictionaryRemoveFields,
    CacheDictionaryRemoveFieldsResponse,
    CacheDictionarySetField,
    CacheDictionarySetFieldResponse,
    CacheDictionarySetFields,
    CacheDictionarySetFieldsResponse,
    CacheGetResponse,
    CacheIncrementResponse,
    CacheListConcatenateBackResponse,
    CacheListConcatenateFrontResponse,
    CacheListFetchResponse,
    CacheListLengthResponse,
    CacheListPopBackResponse,
    CacheListPopFrontResponse,
    CacheListPushBackResponse,
    CacheListPushFrontResponse,
    CacheListRemoveValueResponse,
    CacheSetAddElement,
    CacheSetAddElementResponse,
    CacheSetAddElements,
    CacheSetAddElementsResponse,
    CacheSetFetchResponse,
    CacheSetIfNotExistsResponse,
    CacheSetRemoveElement,
    CacheSetRemoveElementResponse,
    CacheSetRemoveElements,
    CacheSetRemoveElementsResponse,
    CacheSetResponse,
    CreateCacheResponse,
    CreateSigningKeyResponse,
    DeleteCacheResponse,
    ListCachesResponse,
    ListSigningKeysResponse,
    RevokeSigningKeyResponse,
)
from momento.typing import TDictionaryItems


class CacheClientAsync:
    """Async Cache Client.

    Cache and control methods return a response object unique to each request.
    The response object is resolved to a type-safe object of one of several
    sub-types. See the documentation for each response type for details.

    Pattern matching can be used to operate on the appropriate subtype.
    For example, in python 3.10+ if you're deleting a key::

        response = await client.delete(cache_name, key)
        match response:
            case CacheDelete.Success():
                ...they key was deleted or not found...
            case CacheDelete.Error():
                ...there was an error trying to delete the key...

    or equivalently in earlier versions of python::

        response = await client.delete(cache_name, key)
        if isinstance(response, CacheDelete.Success):
            ...
        elif isinstance(response, CacheDelete.Error):
            ...
        else:
            raise Exception("This should never happen")
    """

    _NUM_CLIENTS = 1
    """(async client only) For high load, we might get better performance with multiple clients,
    because the server is configured to allow a max of 100 streams per connection.
`
    In the javascript SDK, multiple clients resulted in an obvious performance improvement.
    However, in the python SDK I have not yet been able to observe a clear benefit.
    So for now, we are putting the plumbing in place so that we can more easily test performance
    with multiple connections in the future, but we are leaving the default value set to 1.

    We are hard-coding the value for now, because we haven't yet designed the API for
    users to use to configure tunables:
    https://github.com/momentohq/dev-eco-issue-tracker/issues/85
    """

    def __init__(self, configuration: Configuration, credential_provider: CredentialProvider, default_ttl: timedelta):
        """Instantiate a client.

        Args:
            configuration (Configuration): An object holding configuration settings for communication with the server.
            credential_provider (CredentialProvider): An object holding the auth token and endpoint information.
            default_ttl (timedelta): A default Time To Live timedelta for cache objects created by this client.
                It is possible to override this setting when calling the set method.

        Raises:
            IllegalArgumentException: If method arguments fail validations.
        Example::

            from datetime import timedelta
            from momento import Configurations, CredentialProvider, CacheClientAsync

            configuration = Configurations.Laptop.latest()
            credential_provider = CredentialProvider.from_environment_variable("MOMENTO_AUTH_TOKEN")
            ttl_seconds = timedelta(seconds=60)
            client = CacheClientAsync(configuration, credential_provider, ttl_seconds)
        """
        _validate_request_timeout(configuration.get_transport_strategy().get_grpc_configuration().get_deadline())
        self._logger = logs.logger
        self._next_client_index = 0
        self._control_client = _ScsControlClient(configuration, credential_provider)
        self._cache_endpoint = credential_provider.cache_endpoint
        self._data_clients = [
            _ScsDataClient(configuration, credential_provider, default_ttl)
            for _ in range(CacheClientAsync._NUM_CLIENTS)
        ]

    async def __aenter__(self) -> CacheClientAsync:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self._control_client.close()
        for data_client in self._data_clients:
            await data_client.close()

    async def create_cache(self, cache_name: str) -> CreateCacheResponse:
        """Creates a cache if it doesn't exist.

        Args:
            cache_name (str): Name of the cache to be created.

        Returns:
            CreateCacheResponse:
        """
        return await self._control_client.create_cache(cache_name)

    async def delete_cache(self, cache_name: str) -> DeleteCacheResponse:
        """Deletes a cache and all of the items within it.

        Args:
            cache_name (str): Name of the cache to be deleted.

        Returns:
            DeleteCacheResponse:
        """
        return await self._control_client.delete_cache(cache_name)

    async def list_caches(self) -> ListCachesResponse:
        """Lists all caches.

        Returns:
            ListCachesResponse:
        """
        return await self._control_client.list_caches()

    async def create_signing_key(self, ttl: timedelta) -> CreateSigningKeyResponse:
        """Creates a Momento signing key.

        Args:
            ttl (timedelta): The key's time-to-live represented as a timedelta

        Returns:
            CreateSigningKeyResponse
        """
        return await self._control_client.create_signing_key(ttl, self._cache_endpoint)

    async def revoke_signing_key(self, key_id: str) -> RevokeSigningKeyResponse:
        """Revokes a Momento signing key, all tokens signed by which will be invalid.

        Args:
            key_id (str): The id of the Momento signing key to revoke

        Returns:
            RevokeSigningKeyResponse
        """
        return await self._control_client.revoke_signing_key(key_id)

    async def list_signing_keys(self) -> ListSigningKeysResponse:
        """Lists all Momento signing keys for the provided auth token.

        Returns:
            ListSigningKeysResponse
        """
        return await self._control_client.list_signing_keys(self._cache_endpoint)

    async def increment(
        self,
        cache_name: str,
        key: str | bytes,
        amount: int = 1,
        ttl: Optional[timedelta] = None,
    ) -> CacheIncrementResponse:
        """Add to a value.

        Args:
            cache_name (str): Name of the cache to store the item in.
            key (str|bytes): The key to set.
            amount (int, optional): The quantity to add to the value. Defaults to 1.
            ttl (Optional[timedelta], optional): TTL for the item in cache.
            This TTL takes precedence over the TTL used when initializing a cache client.
            Defaults to client TTL. If specified must be strictly positive.

        Returns:
            CacheIncrementResponse
        """
        return await self._data_client.increment(cache_name, key, amount, ttl)

    async def set(
        self,
        cache_name: str,
        key: str | bytes,
        value: str | bytes,
        ttl: Optional[timedelta] = None,
    ) -> CacheSetResponse:
        """Set the value in cache with a given time to live (TTL) seconds.

        Args:
            cache_name (str): Name of the cache to store the item in.
            key (str | bytes): The key to set.
            value (str | bytes): The value to be stored.
            ttl (Optional[timedelta], optional): TTL for the item in cache.
            This TTL takes precedence over the TTL used when initializing a cache client.
            Defaults to client TTL. If specified must be strictly positive.

        Returns:
            CacheSetResponse:
        """
        return await self._data_client.set(cache_name, key, value, ttl)

    async def set_if_not_exists(
        self,
        cache_name: str,
        key: str | bytes,
        value: str | bytes,
        ttl: Optional[timedelta] = None,
    ) -> CacheSetIfNotExistsResponse:
        """Like `set`, but it will only set if the key does not already exist.

        Args:
            cache_name (str): Name of the cache to store the item in.
            key (str | bytes): The key to set.
            value (str | bytes): The value to be stored if the key does not exist.
            ttl (Optional[timedelta], optional): TTL for the item in cache.
            This TTL takes precedence over the TTL used when initializing a cache client.
            Defaults to client TTL. If specified must be strictly positive.

        Returns:
            CacheSetIfNotExistsResponse:
        """
        return await self._data_client.set_if_not_exists(cache_name, key, value, ttl)

    async def get(self, cache_name: str, key: str | bytes) -> CacheGetResponse:
        """Get the cache value stored for the given key.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            key (str | bytes): The key to lookup.

        Returns:
            CacheGetResponse:
        """
        return await self._data_client.get(cache_name, key)

    async def delete(self, cache_name: str, key: str | bytes) -> CacheDeleteResponse:
        """Remove the key from the cache.

        Args:
            cache_name (str): Name of the cache to delete the key from.
            key (str | bytes): The key to delete.

        Returns:
            CacheDeleteResponse:
        """
        return await self._data_client.delete(cache_name, key)

    # DICTIONARY COLLECTION METHODS
    async def dictionary_fetch(self, cache_name: str, dictionary_name: str) -> CacheDictionaryFetchResponse:
        """Fetch the entire dictionary from the cache.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): The name of the dictionary to fetch.

        Returns:
            CacheDictionaryFetchResponse: result of the fetch operation and the associated dictionary.
        """
        return await self._data_client.dictionary_fetch(cache_name, dictionary_name)

    async def dictionary_get_field(
        self, cache_name: str, dictionary_name: str, field: str | bytes
    ) -> CacheDictionaryGetFieldResponse:
        """Get the cache value stored for the given dictionary and field.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): The dictionary to lookup.
            field (str | bytes): The field in the dictionary to lookup.

        Returns:
            CacheDictionaryGetFieldResponse: the status and value for the field.
        """
        get_fields_response = await self.dictionary_get_fields(cache_name, dictionary_name, [field])
        if isinstance(get_fields_response, CacheDictionaryGetFields.Hit):
            if len(get_fields_response.responses) == 0:
                return CacheDictionaryGetField.Error(UnknownException("Unknown get fields response had no data"))
            return get_fields_response.responses[0]
        elif isinstance(get_fields_response, CacheDictionaryGetFields.Miss):
            return CacheDictionaryGetField.Miss()
        elif isinstance(get_fields_response, CacheDictionaryGetFields.Error):
            return CacheDictionaryGetField.Error(get_fields_response.inner_exception)
        else:
            return CacheDictionaryGetField.Error(UnknownException(f"Unknown get field response: {get_fields_response}"))

    async def dictionary_get_fields(
        self, cache_name: str, dictionary_name: str, fields: Iterable[str | bytes]
    ) -> CacheDictionaryGetFieldsResponse:
        """Get several values from a dictionary.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): The dictionary to lookup.
            fields (Iterable[str | bytes]): The fields in the dictionary to lookup.

        Returns:
            CacheDictionaryGetFieldsResponse: the status and associated value for each field.
        """
        return await self._data_client.dictionary_get_fields(cache_name, dictionary_name, fields)

    async def dictionary_increment(
        self,
        cache_name: str,
        dictionary_name: str,
        field: str | bytes,
        amount: int = 1,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheDictionaryIncrementResponse:
        """Add an integer quantity to a dictionary value.

        Args:
            cache_name (str): Name of the cache to store the dictionary in.
            dictionary_name (str): Name of the dictionary to increment in.
            field (str | bytes): The field to increment.
            amount (int, optional): The quantity to add to the value. May be positive, negative, or zero. Defaults to 1.
            ttl (CollectionTtl, optional): TTL for the dictionary in cache. This TTL takes precedence over the TTL
                used when initializing a cache client. Defaults to client TTL.
                Defaults to CollectionTtl.from_cache_ttl().

        Returns:
            CacheDictionaryIncrementResponse: result of the increment operation.
        """
        return await self._data_client.dictionary_increment(cache_name, dictionary_name, field, amount, ttl)

    async def dictionary_remove_field(
        self, cache_name: str, dictionary_name: str, field: str | bytes
    ) -> CacheDictionaryRemoveFieldResponse:
        """Remove a field from a dictionary.

        Performs a no-op if the dictionary or field do not exist.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): Name of the dictionary to remove the field from.
            field (str | bytes): Name of the field to remove from the dictionary.

        Returns:
            CacheDictionaryRemoveFieldResponse: result of the remove operation.
        """
        remove_fields_response = await self.dictionary_remove_fields(cache_name, dictionary_name, fields=[field])
        if isinstance(remove_fields_response, CacheDictionaryRemoveFields.Success):
            return CacheDictionaryRemoveField.Success()
        elif isinstance(remove_fields_response, CacheDictionaryRemoveFields.Error):
            return CacheDictionaryRemoveField.Error(remove_fields_response.inner_exception)
        else:
            return CacheDictionaryRemoveField.Error(
                UnknownException(f"Unknown remove fields response: {remove_fields_response}")
            )

    async def dictionary_remove_fields(
        self, cache_name: str, dictionary_name: str, fields: Iterable[str | bytes]
    ) -> CacheDictionaryRemoveFieldsResponse:
        """Remove fields from a dictionary.

        Performs a no-op if the dictionary or a particular field does not exist.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): Name of the dictionary to remove the fields from.
            fields (Iterable[str | bytes]): The fields to remove from the dictionary.

        Returns:
            CacheDictionaryRemoveFieldsResponse: result of the remove fields operation.
        """
        return await self._data_client.dictionary_remove_fields(cache_name, dictionary_name, fields)

    async def dictionary_set_field(
        self,
        cache_name: str,
        dictionary_name: str,
        field: str | bytes,
        value: str | bytes,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheDictionarySetFieldResponse:
        """Set the dictionary field to a value with a given time to live (TTL) seconds.

        Args:
            cache_name (str): Name of the cache to store the dictionary in.
            dictionary_name (str): Name of the dictionary to set.
            field (str | bytes): The field in the dictionary to set.
            value (str | bytes): The value to be stored.
            ttl (CollectionTtl, optional): TTL for the dictionary in cache.
                This TTL takes precedence over the TTL used when initializing a cache client.
                Defaults to CollectionTtl.from_cache_ttl().

        Returns:
            CacheDictionarySetFieldResponse: result of the set operation.
        """
        set_fields_response = await self.dictionary_set_fields(
            cache_name, dictionary_name, items={field: value}, ttl=ttl
        )
        if isinstance(set_fields_response, CacheDictionarySetFields.Success):
            return CacheDictionarySetField.Success()
        elif isinstance(set_fields_response, CacheDictionarySetFields.Error):
            return CacheDictionarySetField.Error(set_fields_response.inner_exception)
        else:
            return CacheDictionarySetField.Error(
                UnknownException(f"Unknown set fields response: {set_fields_response}")
            )

    async def dictionary_set_fields(
        self,
        cache_name: str,
        dictionary_name: str,
        items: TDictionaryItems,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheDictionarySetFieldsResponse:
        """Set several dictionary field-value pairs in the cache.

        Args:
            cache_name (str): Name of the cache to perform the lookup in.
            dictionary_name (str): Name of the dictionary to set.
            items (TDictionaryItems): Field value pairs to store.
            ttl (CollectionTtl, optional): TTL for the dictionary in cache.
                This TTL takes precedence over the TTL used when initializing a cache client.
                Defaults to CollectionTtl.from_cache_ttl().

        Returns:
            CacheDictionarySetFieldsResponse: result of the set fields operation.
        """
        return await self._data_client.dictionary_set_fields(cache_name, dictionary_name, items, ttl)

    # LIST COLLECTION METHODS
    async def list_concatenate_back(
        self,
        cache_name: str,
        list_name: str,
        values: Iterable[str | bytes],
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_front_to_size: Optional[int] = None,
    ) -> CacheListConcatenateBackResponse:
        """Add values to the end of the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to concatenate.
            values: (Iterable[str | bytes]): The values to concatenate.
            ttl: (CollectionTtl, optional): How to treat the list's TTL. Defaults to `CollectionTtl.from_cache_ttl()`
            truncate_front_to_size (Optional[int]): If the list exceeds this size, remove values from
            the start of the list.

        Returns:
            CacheListConcatenateBackResponse:
        """
        return await self._data_client.list_concatenate_back(cache_name, list_name, values, ttl, truncate_front_to_size)

    async def list_concatenate_front(
        self,
        cache_name: str,
        list_name: str,
        values: Iterable[str | bytes],
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_back_to_size: Optional[int] = None,
    ) -> CacheListConcatenateFrontResponse:
        """Add values to the start of the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to concatenate.
            values: (Iterable[str | bytes]): The values to concatenate.
            ttl: (CollectionTtl, optional): How to treat the list's TTL. Defaults to `CollectionTtl.from_cache_ttl()`
            truncate_back_to_size (Optional[int], optional): If the list exceeds this size, remove values from
                the end of the list.

        Returns:
            CacheListConcatenateFrontResponse:
        """
        return await self._data_client.list_concatenate_front(cache_name, list_name, values, ttl, truncate_back_to_size)

    async def list_fetch(self, cache_name: str, list_name: str) -> CacheListFetchResponse:
        """Gets all values from the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to fetch.

        Returns:
            CacheListFetchResponse:
        """
        return await self._data_client.list_fetch(cache_name, list_name)

    async def list_length(self, cache_name: str, list_name: str) -> CacheListLengthResponse:
        """Gets the number of values in the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to fetch.

        Returns:
            CacheListLengthResponse:
        """
        return await self._data_client.list_length(cache_name, list_name)

    async def list_pop_back(self, cache_name: str, list_name: str) -> CacheListPopBackResponse:
        """Gets, removes, and returns the last value from the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to fetch.

        Returns:
            CacheListPopBackResponse:
        """
        return await self._data_client.list_pop_back(cache_name, list_name)

    async def list_pop_front(self, cache_name: str, list_name: str) -> CacheListPopFrontResponse:
        """Gets, removes, and returns the first value from the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to fetch.

        Returns:
            CacheListPopFrontResponse:
        """
        return await self._data_client.list_pop_front(cache_name, list_name)

    async def list_push_back(
        self,
        cache_name: str,
        list_name: str,
        value: str | bytes,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_front_to_size: Optional[int] = None,
    ) -> CacheListPushBackResponse:
        """Add values to the end of the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to push to.
            value: (str | bytes): The value to push.
            ttl: (CollectionTtl, optional): How to treat the list's TTL. Defaults to `CollectionTtl.from_cache_ttl()`
            truncate_front_to_size (Optional[int]): If the list exceeds this size, remove values from
            the start of the list.

        Returns:
            CacheListPushBackResponse:
        """
        return await self._data_client.list_push_back(cache_name, list_name, value, ttl, truncate_front_to_size)

    async def list_push_front(
        self,
        cache_name: str,
        list_name: str,
        value: str | bytes,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
        truncate_back_to_size: Optional[int] = None,
    ) -> CacheListPushFrontResponse:
        """Add values to the start of the list.

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to push to.
            value: (str | bytes): The value to push.
            ttl: (CollectionTtl, optional): How to treat the list's TTL. Defaults to `CollectionTtl.from_cache_ttl()`
            truncate_back_to_size (Optional[int]): If the list exceeds this size, remove values from
                the end of the list.

        Returns:
            CacheListPushFrontResponse:
        """
        return await self._data_client.list_push_front(cache_name, list_name, value, ttl, truncate_back_to_size)

    async def list_remove_value(
        self,
        cache_name: str,
        list_name: str,
        value: str | bytes,
    ) -> CacheListRemoveValueResponse:
        """Removes all matching values from the list.

        Example:
            await client.list_concatenate_front(cache_name, list_name, ['up', 'up', 'down', 'down', 'left', 'right'])
            await client.list_remove_value(cache_name, list_name, 'up')
            fetch_resp = client.list_fetch(cache_name, list_name)

            # ['down', 'down', 'left', 'right']
            print(fetch_resp.values_string)

        Args:
            cache_name (str): The cache where the list is.
            list_name (str): The name of the list to remove values from.
            value: (str | bytes): The value to remove.

        Returns:
            CacheListRemoveValueResponse
        """
        return await self._data_client.list_remove_value(cache_name, list_name, value)

    # SET COLLECTION METHODS
    async def set_add_element(
        self,
        cache_name: str,
        set_name: str,
        element: str | bytes,
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheSetAddElementResponse:
        """Adds an element to a set.

        Args:
            cache_name (str): The cache name with the set.
            set_name (str): The name of the set to add to.
            element (str | bytes): The element to add.
            ttl: (CollectionTtl, optional): How to treat the set's TTL. Defaults to `CollectionTtl.from_cache_ttl()`

        Returns:
            CacheSetAddElementResponse
        """
        resp = await self.set_add_elements(cache_name, set_name, (element,), ttl=ttl)

        if isinstance(resp, CacheSetAddElements.Success):
            return CacheSetAddElement.Success()
        elif isinstance(resp, CacheSetAddElements.Error):
            return CacheSetAddElement.Error(resp.inner_exception)
        else:
            raise UnknownException(f"Unknown response type: {type(resp)}")

    async def set_add_elements(
        self,
        cache_name: str,
        set_name: str,
        elements: Iterable[str | bytes],
        *,
        ttl: CollectionTtl = CollectionTtl.from_cache_ttl(),
    ) -> CacheSetAddElementsResponse:
        """Add an elements to a set.

        Args:
            cache_name (str): The cache name with the set.
            set_name (str): The name of the set to add to.
            elements (Iterable[str | bytes]): The element to add.
            ttl: (CollectionTtl, optional): How to treat the set's TTL. Defaults to `CollectionTtl.from_cache_ttl()`

        Returns:
            CacheSetAddElementsResponse
        """
        return await self._data_client.set_add_elements(cache_name, set_name, elements, ttl)

    async def set_fetch(
        self,
        cache_name: str,
        set_name: str,
    ) -> CacheSetFetchResponse:
        """Fetches a set.

        Args:
            cache_name (str): The cache name with the set.
            set_name (str): The name of the set to fetch.

        Returns:
            CacheSetFetchResponse
        """
        return await self._data_client.set_fetch(cache_name, set_name)

    async def set_remove_element(
        self, cache_name: str, set_name: str, element: str | bytes
    ) -> CacheSetRemoveElementResponse:
        """Remove an element from a set.

        Args:
            cache_name (str): The cache name with the set.
            set_name (str): The name of the set to remove from.
            element (str | bytes): The element to remove.

        Returns:
            CacheSetRemoveElementResponse
        """
        resp = await self.set_remove_elements(cache_name, set_name, (element,))
        if isinstance(resp, CacheSetRemoveElements.Success):
            return CacheSetRemoveElement.Success()
        elif isinstance(resp, CacheSetRemoveElements.Error):
            return CacheSetRemoveElement.Error(resp.inner_exception)
        else:
            raise UnknownException(f"Unknown response type: {type(resp)}")

    async def set_remove_elements(
        self, cache_name: str, set_name: str, elements: Iterable[str | bytes]
    ) -> CacheSetRemoveElementsResponse:
        """Remove elements from a set.

        Args:
            cache_name (str): The cache name with the set.
            set_name (str): The name of the set to remove from.
            elements (Iterable[str | bytes]): The element to remove.

        Returns:
            CacheSetRemoveElementsResponse
        """
        return await self._data_client.set_remove_elements(cache_name, set_name, elements)

    @property
    def _data_client(self) -> _ScsDataClient:
        client = self._data_clients[self._next_client_index]
        self._next_client_index = (self._next_client_index + 1) % len(self._data_clients)
        return client



from __future__ import annotations

import json
from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from typing import Any

from momento.responses.response import ControlResponse

from ...mixins import ErrorResponseMixin


class CreateSigningKeyResponse(ControlResponse):
    """Parent response type for a cache `create_signing_key` request.

    Its subtypes are:
    - `CreateSigningKey.Success`
    - `CreateSigningKey.Error`

    See `CacheClient` for how to work with responses.
    """


class CreateSigningKey(ABC):
    """Groups all `CreateSigningKeyResponse` derived types under a common namespace."""

    @dataclass
    class Success(CreateSigningKeyResponse):
        """The response from creating a signing key."""

        key_id: str
        """The ID of the signing key"""
        endpoint: str
        """The endpoint of the signing key"""
        key: str
        """The signing key as a JSON string"""
        expires_at: datetime
        """When the key expires"""

        @staticmethod
        def from_grpc_response(grpc_create_signing_key_response: Any, endpoint: str) -> CreateSigningKey.Success:  # type: ignore[misc] # noqa: E501
            key: str = grpc_create_signing_key_response.key
            key_id: str = json.loads(key)["kid"]
            expires_at: datetime = datetime.fromtimestamp(grpc_create_signing_key_response.expires_at)
            return CreateSigningKey.Success(key_id, endpoint, key, expires_at)

    class Error(CreateSigningKeyResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

import asyncio
import logging
from typing import Callable

import grpc

from momento.retry import RetryableProps, RetryStrategy

# TODO: This is very duplicative of the synchronous retry interceptor; we need to
# DRY these up, but for now I am prioritizing getting a fix out for a customer.

logger = logging.getLogger("retry-interceptor")


# TODO: We need to send retry count information to the server so that we
# will have some visibility into how often this is happening to customers:
# https://github.com/momentohq/client-sdk-javascript/issues/80
# TODO: we need to add backoff/jitter for the retries:
# https://github.com/momentohq/client-sdk-javascript/issues/81


class RetryInterceptor(grpc.aio.UnaryUnaryClientInterceptor):
    def __init__(self, retry_strategy: RetryStrategy):
        self._retry_strategy = retry_strategy

    async def intercept_unary_unary(
        self,
        continuation: Callable[
            [grpc.aio._interceptor.ClientCallDetails, grpc.aio._typing.RequestType],
            grpc.aio._call.UnaryUnaryCall,
        ],
        client_call_details: grpc.aio._interceptor.ClientCallDetails,
        request: grpc.aio._typing.RequestType,
    ) -> grpc.aio._call.UnaryUnaryCall | grpc.aio._typing.ResponseType:
        attempt_number = 1
        while True:
            call = await continuation(client_call_details, request)
            response_code = await call.code()

            retryTime = self._retry_strategy.determine_when_to_retry(
                RetryableProps(response_code, client_call_details.method.decode("utf-8"), attempt_number)
            )

            if retryTime is None:
                return call

            attempt_number += 1
            await asyncio.sleep(retryTime)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['momento',
 'momento.auth',
 'momento.config',
 'momento.config.transport',
 'momento.errors',
 'momento.internal',
 'momento.internal._utilities',
 'momento.internal.aio',
 'momento.internal.synchronous',
 'momento.requests',
 'momento.responses',
 'momento.responses.control',
 'momento.responses.control.cache',
 'momento.responses.control.signing_key',
 'momento.responses.data',
 'momento.responses.data.dictionary',
 'momento.responses.data.list',
 'momento.responses.data.scalar',
 'momento.responses.data.set',
 'momento.retry']

package_data = \
{'': ['*']}

install_requires = \
['grpcio>=1.50.0,<2.0.0',
 'momento-wire-types>=0.39,<0.40',
 'pyjwt>=2.4.0,<3.0.0']

setup_kwargs = {
    'name': 'momento',
    'version': '1.0.0',
    'description': 'SDK for Momento',
    'long_description': '<head>\n  <meta name="Momento Python Client Library Documentation" content="Python client software development kit for Momento Serverless Cache">\n</head>\n<img src="https://docs.momentohq.com/img/logo.svg" alt="logo" width="400"/>\n\n[![project status](https://momentohq.github.io/standards-and-practices/badges/project-status-official.svg)](https://github.com/momentohq/standards-and-practices/blob/main/docs/momento-on-github.md)\n[![project stability](https://momentohq.github.io/standards-and-practices/badges/project-stability-beta.svg)](https://github.com/momentohq/standards-and-practices/blob/main/docs/momento-on-github.md) \n\n# Momento Python Client Library\n\n\nPython client SDK for Momento Serverless Cache: a fast, simple, pay-as-you-go caching solution without\nany of the operational overhead required by traditional caching solutions!\n\n\n\n## Getting Started :running:\n\n### Requirements\n\n- [Python 3.7](https://www.python.org/downloads/) or above is required\n- A Momento Auth Token is required, you can generate one using the [Momento CLI](https://github.com/momentohq/momento-cli)\n\n### Examples\n\nReady to dive right in? Just check out the [examples](./examples/README.md) directory for complete, working examples of\nhow to use the SDK.\n\n### Installation\n\nThe [Momento SDK is available on PyPi](https://pypi.org/project/momento/). To install via pip:\n\n```bash\npip install momento\n```\n\n### Usage\n\nThe examples below require an environment variable named MOMENTO_AUTH_TOKEN which must\nbe set to a valid [Momento authentication token](https://docs.momentohq.com/docs/getting-started#obtain-an-auth-token).\n\nPython 3.10 introduced the `match` statement, which allows for [structural pattern matching on objects](https://peps.python.org/pep-0636/#adding-a-ui-matching-objects).\nIf you are running python 3.10 or greater, here is a quickstart you can use in your own project:\n\n```python\nfrom datetime import timedelta\n\nfrom momento import CacheClient, Configurations, CredentialProvider\nfrom momento.responses import CacheGet, CacheSet, CreateCache\n\nif __name__ == "__main__":\n    cache_name = "default-cache"\n    with CacheClient(\n        configuration=Configurations.Laptop.v1(),\n        credential_provider=CredentialProvider.from_environment_variable("MOMENTO_AUTH_TOKEN"),\n        default_ttl=timedelta(seconds=60),\n    ) as cache_client:\n        create_cache_response = cache_client.create_cache(cache_name)\n        match create_cache_response:\n            case CreateCache.CacheAlreadyExists():\n                print(f"Cache with name: {cache_name} already exists.")\n            case CreateCache.Error() as error:\n                raise error.inner_exception\n\n        print("Setting Key: foo to Value: FOO")\n        set_response = cache_client.set(cache_name, "foo", "FOO")\n        match set_response:\n            case CacheSet.Error() as error:\n                raise error.inner_exception\n\n        print("Getting Key: foo")\n        get_response = cache_client.get(cache_name, "foo")\n        match get_response:\n            case CacheGet.Hit() as hit:\n                print(f"Look up resulted in a hit: {hit}")\n                print(f"Looked up Value: {hit.value_string!r}")\n            case CacheGet.Miss():\n                print("Look up resulted in a: miss. This is unexpected.")\n            case CacheGet.Error() as error:\n                raise error.inner_exception\n\n```\n\nThe above code uses [structural pattern matching](https://peps.python.org/pep-0636/), a feature introduced in Python 3.10.\nUsing a Python version less than 3.10? No problem. Here is the same example compatible across all versions of Python:\n\n```python\nfrom datetime import timedelta\nfrom momento import CacheClient, Configurations, CredentialProvider\nfrom momento.responses import CacheGet, CacheSet, CreateCache\n\nif __name__ == "__main__":\n    cache_name = \'default-cache\'\n    with CacheClient(configuration=Configurations.Laptop.v1(),\n                     credential_provider=CredentialProvider.from_environment_variable(\'MOMENTO_AUTH_TOKEN\'),\n                     default_ttl=timedelta(seconds=60)\n                     ) as cache_client:\n        create_cache_response = cache_client.create_cache(cache_name)\n        if isinstance(create_cache_response, CreateCache.CacheAlreadyExists):\n            print(f"Cache with name: {cache_name} already exists.")\n        elif isinstance(create_cache_response, CreateCache.Error):\n            raise create_cache_response.inner_exception\n\n        print("Setting Key: foo to Value: FOO")\n        set_response = cache_client.set(cache_name, \'foo\', \'FOO\')\n        if isinstance(set_response, CacheSet.Error):\n            raise set_response.inner_exception\n\n        print("Getting Key: foo")\n        get_response = cache_client.get(cache_name, \'foo\')\n        if isinstance(get_response, CacheGet.Hit):\n            print(f"Look up resulted in a hit: {get_response.value_string}")\n            print(f"Looked up Value: {get_response.value_string}")\n        elif isinstance(get_response, CacheGet.Miss):\n            print("Look up resulted in a: miss. This is unexpected.")\n        elif isinstance(get_response, CacheGet.Error):\n            raise get_response.inner_exception\n```\n\n### Logging\n\nTo avoid cluttering DEBUG logging with per-method logs the Momento SDK adds a TRACE logging level. This will only happen\nif the TRACE level does not already exist.\n\nTo enable TRACE level logging you can call logging.basicConfig() before making any log statements:\n\n```python\nimport logging\n\nlogging.basicConfig(level=\'TRACE\')\n```\n\n### Error Handling\n\nComing Soon!\n\n### Tuning\n\nComing Soon!\n\n----------------------------------------------------------------------------------------\nFor more info, visit our website at [https://gomomento.com](https://gomomento.com)!\n',
    'author': 'Momento',
    'author_email': 'hello@momentohq.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gomomento.com',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.12',
}


setup(**setup_kwargs)
